from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class BytearrayTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = bytearray([1,2,3])
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = bytearray([1,2,3])
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_contains(self):
        self.assertCodeExecution("""
            print(bytearray([1,2,3]) in bytearray([1,2]))
            print(bytearray([1,2]) in bytearray([1,2,3]))
            print(bytearray([1,2,4]) in bytearray([1,2,3]))
            print(bytearray([8,9,0,1]) in bytearray([1,2,3]))
            print(101 in bytearray([1,2,3]))
            print(101 in bytearray([1,2,3,101]))
            print(b'pybee' in bytearray([1,2]))
            print(bytearray([1,2]) in b'pybee')
        """)
        self.assertCodeExecution("""
            try:
                print(300 in bytearray([1,2,3]))
                print("No error raised")
            except ValueError:
                print("Raised a ValueError")
        """)
        self.assertCodeExecution("""
            try:
                print(['b', 'e'] in bytearray([1,2,3]))
                print("No error raised")
            except TypeError:
                print("Raised a TypeError")
        """)

    def test_capitalize(self):
        self.assertCodeExecution("""
            print(bytearray(b'abc').capitalize())
            print(bytearray().capitalize())
        """)

    def test_islower(self):
        # TODO: add this test when adding support for literal hex bytes
        # print(b'\xf0'.islower())

        self.assertCodeExecution("""
            print(bytearray(b'abc').islower())
            print(bytearray(b'').islower())
            print(bytearray(b'Abccc').islower())
            print(bytearray(b'HELLO WORD').islower())
            print(bytearray(b'@#$%!').islower())
            print(bytearray(b'hello world').islower())
            print(bytearray(b'hello world   ').islower())
        """)

    def test_isspace(self):
        self.assertCodeExecution("""
            print(bytearray(b'testupper').isspace())
            print(bytearray(b'test isspace').isspace())
            print(bytearray(b' ').isspace())
            print(bytearray(b'').isspace())
            print(bytearray(b'\x46').isspace())
            print(bytearray(b'   \t\t').isspace())
            print(bytearray(b' \x0b').isspace())
            print(bytearray(b' \f').isspace())
            print(bytearray(b' \\n').isspace())
            print(bytearray(b' \\r').isspace())
        """)

    def test_upper(self):
        # TODO: add this test when adding support for literal hex bytes
        # print(bytearray(b'\xf0').upper())

        self.assertCodeExecution("""
            print(bytearray(b'abc').upper())
            print(bytearray(b'').upper())
            print(bytearray(b'Abccc').upper())
            print(bytearray(b'HELLO WORD').upper())
            print(bytearray(b'@#$%!').upper())
            print(bytearray(b'hello world').upper())
            print(bytearray(b'hello world   ').upper())
        """)

    def test_ljust(self):
        self.assertCodeExecution("""
            print(bytearray(b'testMoreThanWidth').ljust(5))
            print(bytearray(b'testEqualWidth').ljust(14))
            print(bytearray(b'testLessThanWidth').ljust(20))
            print(bytearray(b'testMoreWithFill').ljust(2, b'x'))
            print(bytearray(b'testEqualWithFill').ljust(17, b'x'))
            print(bytearray(b'testLessWithFill').ljust(25, b'x'))
            print(bytearray(b'testNegative').ljust(-20))
            print(bytearray(b'').ljust(5))
            print(bytearray(b'testNoChangeWidthOne').ljust(True, b'x'))
            print(bytearray(b'testBArraySecondArg').ljust(True, bytearray(b'x')))
            try:
                print(bytearray(b'testStrArgError').ljust('5'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testMoreLengthError').ljust(12, b'as'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testStrFillingChar').ljust(12, 'a'))
            except Exception as e:
                print(str(e))
        """)

    def test_rjust(self):
        self.assertCodeExecution("""
            print(bytearray(b'testMoreThanWidth').rjust(5))
            print(bytearray(b'testEqualWidth').rjust(14))
            print(bytearray(b'testLessThanWidth').rjust(20))
            print(bytearray(b'testMoreWithFill').rjust(2, b'x'))
            print(bytearray(b'testEqualWithFill').rjust(17, b'x'))
            print(bytearray(b'testLessWithFill').rjust(25, b'x'))
            print(bytearray(b'testNegative').rjust(-20))
            print(bytearray(b'').rjust(5))
            print(bytearray(b'testNoChangeWidthOne').rjust(True, b'x'))
            print(bytearray(b'testBArraySecondArg').rjust(True, bytearray(b'x')))
            try:
                print(bytearray(b'testStrArgError').rjust('5'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testMoreLengthError').rjust(12, b'as'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testStrFillingChar').rjust(12, 'a'))
            except Exception as e:
                print(str(e))
        """)

    def test_expandtabs(self):
        self.assertCodeExecution("""
            print(bytearray(b'testNoTabs').expandtabs())
            print(bytearray(b'test\t').expandtabs())
            print(bytearray(b'testDoubleTab\t\t').expandtabs())
            print(bytearray(b'testTab\tandText').expandtabs())
            print(bytearray(b'testTab\t').expandtabs(4))
            print(bytearray(b'testTab\t\t').expandtabs(4))
            print(bytearray(b'test\t\t').expandtabs(-4))
            print(bytearray(b'').expandtabs(5))
            try:
                print(bytearray(b'testErrorChar\t').expandtabs('a'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testErrorChars\t').expandtabs('as'))
            except Exception as e:
                print(str(e))
            try:
                print(bytearray(b'testErrorCharNum\t').expandtabs('1'))
            except Exception as e:
                print(str(e))
        """)

    def test_isalpha(self):
        # TODO: add this test when adding support for literal hex bytes
        # print(bytearray(b'\xf0').isalpha())

        self.assertCodeExecution("""
            print(bytearray(b'abc').isalpha())
            print(bytearray(b'').isalpha())
            print(bytearray(b'Abccc').isalpha())
            print(bytearray(b'HELLO WORD').isalpha())
            print(bytearray(b'@#$%!').isalpha())
            print(bytearray(b'hello world').isalpha())
            print(bytearray(b'hello world   ').isalpha())
        """)

    def test_isupper(self):
        self.assertCodeExecution("""
            print(bytearray(b'abc').isupper())
            print(bytearray(b'ABC').isupper())
            print(bytearray(b'').isupper())
            print(bytearray(b'Abccc').isupper())
            print(bytearray(b'HELLO WORD').isupper())
            print(bytearray(b'@#$%!').isupper())
            print(bytearray(b'hello world').isupper())
            print(bytearray(b'hello world   ').isupper())
        """)

    def test_lower(self):
        self.assertCodeExecution("""
            print(bytearray(b"abc").lower())
            print(bytearray(b"HELLO WORLD!").lower())
            print(bytearray(b"hElLO wOrLd").lower())
            print(bytearray(b"[Hello] World").lower())
            """)

    def test_count(self):
        self.assertCodeExecution("""
            print(bytearray(b'abcabca').count(97))
            print(bytearray(b'abcabca').count(b'abc'))
            print(bytearray(b'qqq').count(b'q'))
            print(bytearray(b'qqq').count(b'qq'))
            print(bytearray(b'qqq').count(b'qqq'))
            print(bytearray(b'qqq').count(b'qqqq'))
            print(bytearray(b'abcdefgh').count(b'bc',-7, -5))
            print(bytearray(b'abcdefgh').count(b'bc',1, -5))
            print(bytearray(b'abcdefgh').count(b'bc',0, 3))
            print(bytearray(b'abcdefgh').count(b'bc',-7, 500))
            print(bytearray(b'qqaqqbqqqcqqqdqqqqeqqqqf').count(b'qq'),1)
            print(bytearray(b'').count(b'q'),0)
        """)

    def test_find(self):
        self.assertCodeExecution("""
            print(bytearray(b'').find(b'a'))
            print(bytearray(b'abcd').find(b''))
            print(bytearray(b'abcd').find(b'...'))
            print(bytearray(b'abcd').find(b'a'))
            print(bytearray(b'abcd').find(b'b'))
            print(bytearray(b'abcd').find(b'c'))
            print(bytearray(b'abcd').find(b'd'))
            print(bytearray(b'abcd').find(bytearray(b'ab')))
            print(bytearray(b'abcd').find(b'bc'))
            print(bytearray(b'abcd').find(b'cd'))
            print(bytearray(b'abcd').find(b'cd', 2))
            print(bytearray(b'abcd').find(bytearray(b'ab'), 3))
            print(bytearray(b'abcd').find(b'cd', 2, 3))
            print(bytearray(b'abcd').find(bytearray(b'ab'), 3, 4))
        """)

    def test_center(self):
        self.assertCodeExecution("""
            print(bytearray(b'pybee').center(12))
            print(bytearray(b'pybee').center(13))
            print(bytearray(b'pybee').center(2))
            print(bytearray(b'pybee').center(2, b'a'))
            print(bytearray(b'pybee').center(12, b'a'))
            print(bytearray(b'pybee').center(13, b'a'))
            print(bytearray(b'pybee').center(-5))
            print(bytearray(b'').center(5))
            print(bytearray(b'pybee').center(True, b'a'))
            print(bytearray(b'pybee').center(True, bytearray(b'a')))
        """)

    def test_title(self):
        self.assertCodeExecution(r"""
            print(bytearray(b"").title())
            print(bytearray(b"abcd").title())
            print(bytearray(b"NOT").title())
            print(bytearray(b"coca cola").title())
            print(bytearray(b"they are from UK, are they not?").title())
            print(bytearray(b'/@.').title())
            print(bytearray(b'\x46\x55\x43\x4B').title())
            print(bytearray(b"py.bee").title())
        """)

    def test_istitle(self):
        self.assertCodeExecution(r"""
            print(bytearray(b"").istitle())
            print(bytearray(b"abcd").istitle())
            print(bytearray(b"NOT").istitle())
            print(bytearray(b"coca cola").istitle())
            print(bytearray(b"they are from UK, are they not?").istitle())
            print(bytearray(b'/@.').istitle())
            print(bytearray(b'\x46\x55\x43\x4B').istitle())
            print(bytearray(b"py.bee").title())
        """)

    def test_repr(self):
        self.assertCodeExecution(r"""
            print(repr(bytearray(b'\xc8')))
            print(repr(bytearray(b'abcdef \xc8 abcdef')))
            print(repr(bytearray(b'abcdef \xc8 abcdef\n\r\t')))
            print(bytearray(b'abcdef \xc8 abcdef\n\r\t'))
            for b in range(0, 256, 16):
                print(repr(bytearray(range(b, b+16))))
            for b in range(0, 256, 16):
                print(bytearray(range(b, b+16)))
        """)

    def test_endswith(self):
        self.assertCodeExecution(r"""
            print(bytearray(b'banana').endswith(b'ana'))
            print(bytearray(b'banana').endswith(b''))
            print(bytearray(b'').endswith(b'ana'))
            print(bytearray(b'').endswith(b''))
        """)

    def test_startswith(self):
        self.assertCodeExecution(r"""
            print(bytearray(b'banana').startswith(b'ana'))
            print(bytearray(b'banana').startswith(b''))
            print(bytearray(b'').startswith(b'ana'))
            print(bytearray(b'').startswith(b''))
        """)

    def test_isalnum(self):
        self.assertCodeExecution("""
            print(bytearray(b'0').isalnum())
            print(bytearray(b'9').isalnum())
            print(bytearray(b'1234567890').isalnum())
            print(bytearray(b'89A23gM23z').isalnum())
            print(bytearray(b':923').isalnum())
            print(bytearray(b'\\923').isalnum())
            print(bytearray(b' jdf fhd 33').isalnum())
            print(bytearray(b'@#$%^&*').isalnum())
            print(bytearray(b'"478\t47ads:').isalnum())
            print(bytearray(b'AbZ').isalnum())
        """)

    def test_isdigit(self):
        self.assertCodeExecution("""
            print(bytearray(b'0').isdigit())
            print(bytearray(b'9').isdigit())
            print(bytearray(b'1234567890').isdigit())
            print(bytearray(b'8923g23823').isdigit())
            print(bytearray(b'923').isdigit())
            print(bytearray(b'\\923').isdigit())
            print(bytearray(b'000').isdigit())
            print(bytearray(b'@#$%^&*').isdigit())
            print(bytearray(b'"478\t47ads:').isdigit())
            print(bytearray(b'AbZ').isdigit())
        """)

    def test_join(self):
        self.assertCodeExecution("""
            b = bytearray(b'.')
            print(b.join([b'12', b'dh']))
            print(b.join([bytearray(b'12'), bytearray(b'dh')]))
            b = bytearray(b' ')
            print(b.join([b'd', bytearray(b'l22-'), b'=ej*']))
            print(b.join([bytearray(b'31'), b'`', b'^']))
            print(b.join([bytearray(b'dh')]))
            b = bytearray(b'%#@!')
            print(b.join([b'1',b'd',b'<']))
            print(b.join([b'12']))
        """)


class UnaryBytearrayOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'bytearray'


class BinaryBytearrayOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'bytearray'

    not_implemented_versions = {
        'test_modulo_complex': (3.4, ),
    }


class InplaceBytearrayOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'bytearray'

    not_implemented_versions = {
        'test_modulo_complex': (3.4, ),
    }



from unittest import expectedFailure

from ..utils import TranspileTestCase


class ExceptionTests(TranspileTestCase):

    def test_raise(self):
        # Caught exception
        self.assertCodeExecution("""
            raise KeyError("This is the name")
            print('Done.')
            """, exits_early=True)

    def test_raise_without_any_params(self):
        self.assertCodeExecution("""
            raise ValueError()
            print('Done.')
        """, exits_early=True)

    def test_raise_by_classname(self):
        self.assertCodeExecution("""
            raise ValueError
            print('Done.')
        """, exits_early=True)

    @expectedFailure
    def test_raise_existing_error(self):
        self.assertCodeExecution("""
            error1 = ValueError
            error2 = ValueError()
            error3 = ValueError("This is the name")

            try:
                raise error1
            except ValueError:
                print("Done")

            try:
                raise error2
            except ValueError:
                print("Done")

            try:
                raise error3
            except ValueError:
                print("Done")
        """)

    def test_raise_catch(self):
        self.assertCodeExecution("""
            try:
                raise KeyError("This is the name")
                print("No error")
            except KeyError:
                print("Got a Key Error")
            print('Done.')
            """)

    def test_reraise(self):
        self.assertCodeExecution("""
            try:
                raise KeyError("This is the name")
                print("No error")
            except KeyError:
                print("Got error")
                raise
            print('Done.')
            """, exits_early=True)

    def test_reraise_anonymous(self):
        self.assertCodeExecution("""
            try:
                raise KeyError("This is the name")
                print("No error")
            except:
                print("Got error")
                raise
            print('Done.')
            """, exits_early=True)

    def test_reraise_named(self):
        self.assertCodeExecution("""
            try:
                raise KeyError("This is the name")
                print("No error")
            except KeyError as e:
                print("Got error")
                raise
            print('Done.')
            """, exits_early=True)

    def test_raise_custom_exception(self):
        self.assertCodeExecution("""
            class MyException(Exception):
                pass

            try:
                raise MyException()
            except MyException:
                print("Got a custom exception")
            print('Done.')
            """)

    def test_raising_exceptions_multiple_args(self):
        self.assertCodeExecution("""
            for exc in [KeyError, ValueError, TypeError]:
                try:
                    raise exc("one")
                except Exception as e:
                    print(e, e.args, str(e), repr(e))
                try:
                    raise exc("one", 2)
                except Exception as e:
                    print(e, e.args)
        """)

    @expectedFailure
    def test_raise_custom_exception_import_from(self):
        self.assertCodeExecution(
            """
            from example import *

            try:
                raise MyException("This is the exception")
            except MyException:
                print("Got a custom exception")
            print('Done.')
            """,
            extra_code={
                'example':
                    """
                    class MyException(Exception):
                        pass

                    """
            })

    @expectedFailure
    def test_stopiteration_equality(self):
        # This is kwown and StopIteration is a singleton by design.
        # See org/python/exceptions/StopIteration
        self.assertCodeExecution("""
            x = iter([])
            y = iter([])

            try:
                next(x)
            except StopIteration as e1:
                try:
                    next(y)
                except StopIteration as e2:
                    print(e1 == e2)
        """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase

from unittest import expectedFailure


class VarsTests(TranspileTestCase):
    @expectedFailure
    def test_simple(self):
        self.assertCodeExecution("""
            print("There are %s vars" % len(vars()))
            x = 1
            y = 'z'
            print("There are %s vars" % len(vars()))

            def method():
                print("In method: there are %s vars" % len(vars()))

                print("vars()['x'] =", vars()['x'])
                print("vars()['y'] =", vars()['y'])
                try:
                    print("vars()['z'] =", vars()['z'])
                except KeyError:
                    print("Variable z not defined")

                vars()[y] = 2

                print("In method: there are %s vars" % len(vars()))

            method()

            print("There are %s vars" % len(vars()))
            print("vars()['x'] =", vars()['x'])
            print("vars()['y'] =", vars()['y'])
            print("vars()['z'] =", vars()['z'])
            print('x', x)
            print('y', y)
            print('z', z)

            print('Done')
        """, run_in_function=False)

    @expectedFailure
    def test_preloaded_vars(self):
        self.assertCodeExecution("""
            v = vars
            print("There are %s vars" % len(v()))

            x = 1
            y = 'z'
            print("There are %s vars" % len(v()))

            def method():
                print("There are %s vars" % len(v()))
                x = 1
                y = 'z'
                print("There are %s vars" % len(v()))

            method()
        """, run_in_function=False)


class BuiltinVarsFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["vars"]

    not_implemented = [
        'test_bool',
        'test_bytearray',
        'test_bytes',
        'test_class',
        'test_complex',
        'test_dict',
        'test_float',
        'test_frozenset',
        'test_int',
        'test_list',
        'test_None',
        'test_NotImplemented',
        'test_set',
        'test_str',
        'test_tuple',
        'test_range',
        'test_slice',
        'test_obj',
    ]



from ..utils import TranspileTestCase

from unittest import expectedFailure


class ComparisonTests(TranspileTestCase):
    def test_bool(self):
        self.assertCodeExecution("""
            x = 1234
            if x:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = object()
            if x:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = None
            if x:
                print('True')
            else:
                print('False')
            """)

    def test_is(self):
        self.assertCodeExecution("""
            x = 1234
            if x is 1234:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1234
            if x is 2345:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = None
            if x is None:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1234
            if x is None:
                print('True')
            else:
                print('False')
            """)

    def test_is_not(self):
        self.assertCodeExecution("""
            x = 1234
            if x is not 2345:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1234
            if x is not 1234:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1234
            if x is not None:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = None
            if x is not None:
                print('True')
            else:
                print('False')
            """)

    def test_lt(self):
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 5
            if x < 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('True')
            else:
                print('False')
            """)

    def test_le(self):
        self.assertCodeExecution("""
            x = 1
            if x <= 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 5
            if x <= 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 10
            if x <= 5:
                print('True')
            else:
                print('False')
            """)

    def test_gt(self):
        self.assertCodeExecution("""
            x = 10
            if x > 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 5
            if x > 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1
            if x > 5:
                print('True')
            else:
                print('False')
            """)

    def test_ge(self):
        self.assertCodeExecution("""
            x = 10
            if x >= 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 5
            if x >= 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 1
            if x >= 5:
                print('True')
            else:
                print('False')
            """)

    def test_eq(self):
        self.assertCodeExecution("""
            x = 10
            if x == 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 5
            if x == 5:
                print('True')
            else:
                print('False')
            """)

    def test_ne(self):
        self.assertCodeExecution("""
            x = 5
            if x != 5:
                print('True')
            else:
                print('False')
            """)

        self.assertCodeExecution("""
            x = 10
            if x != 5:
                print('True')
            else:
                print('False')
            """)

    def test_eq_empty_class(self):
        self.assertCodeExecution("""
            class A:
                pass
            x = A()
            y = A()
            # fallback to object identity
            print(x == x)
            print(x == y)
            print(x.__eq__(x))
            print(x.__eq__(y))
            print(y.__eq__(x))
            """)

    def test_ne_empty_class(self):
        self.assertCodeExecution("""
            class A:
                pass
            x = A()
            y = A()
            print(x != x)
            print(x != y)
            print(x.__ne__(x))
            print(x.__ne__(y))
            print(y.__ne__(x))
            """)

    def test_eq_non_bool(self):
        self.assertCodeExecution("""
            class A:
                def __init__(self, num):
                    self.num = num
                def __eq__(self, other):
                    return self.num
            x = A(0)
            y = A(1)
            print(x == x) # 0
            print(x == y) # __eq__() returns 0 => False
            print(y == x) # __eq__() returns 1 => True
            print(x.__eq__(x)) # 0
            print(x.__eq__(y)) # 0
            print(y.__eq__(x)) # 1

            # test evaluation by ==
            if (x == x):
                print('A')
            if (x == y):
                print('B')
            if (y == x):
                print('C')
            """)

    def test_ne_non_bool(self):
        self.assertCodeExecution("""
            class A:
                def __init__(self, num):
                    self.num = num
                def __ne__(self, other):
                    return self.num
            x = A(0)
            y = A(1)
            print(x != x) # 0
            print(x != y) # __eq__() returns 0 => False
            print(y != x) # __eq__() returns 1 => True
            print(x.__ne__(x)) # 0
            print(x.__ne__(y)) # 0
            print(y.__ne__(x)) # 1

            # test evaluation by !=
            if (x != x):
                print('A')
            if (x != y):
                print('B')
            if (y != x):
                print('C')
            """)

    def test_eq_ne_notimplemented(self):
        self.assertCodeExecution("""
            class A:
                def __eq__(self, other):
                    return NotImplemented
                def __ne__(self, other):
                    return NotImplemented
            x = A()

            # == and != are special, should fallback to testing for object equality
            # instead of TypeError
            print(x == x) # True, same object
            print(x != x) # False, same object
            """)

    def test_eq_within_datatypes(self):
        self.assertCodeExecution("""
            class F:
                def __init__(self, v):
                    self.v = v
                def __eq__(self, other):
                    return self.v == other.v + 1
            k = F(5)
            l = F(6)

            print(k == k) # False
            print([k] == [k]) # True as identity tested first before trying __eq__()
            print({'hi': k} == {'hi': k})
            print((k,) == (k,))

            print(k == l) # False as 5 == 6 + 1
            print(l == k) # True as 6 == 5 + 1
            print([l] == [k]) # True
            print({'hi': l} == {'hi': k}) # True
            print((l,) == (k,))
            """)

    def test_eq_reflected_within_datatypes(self):
        self.assertCodeExecution("""
            class F:
                def __init__(self, v):
                    self.v = v
                def __eq__(self, other):
                    return NotImplemented
            class G:
                def __init__(self, v):
                    self.v = v
                def __eq__(self, other):
                    return self.v == other.v + 1
            f = F(5)
            g = G(6)

            print([f] == [g]) # True (reflected)
            print([g] == [f]) # True (not reflected)
            print({'hi': f} == {'hi': g})
            print({'hi': g} == {'hi': f})
            print((f,) == (g,))
            print((g,) == (f,))
            """)

    def test_ne_within_datatypes(self):
        self.assertCodeExecution("""
            class F:
                def __init__(self, v):
                    self.v = v
                def __ne__(self, other):
                    return self.v == other.v + 1
            k = F(5)
            l = F(6)

            print(k != k) # False
            print([k] != [k]) # True as identity tested first before trying __eq__()
            print({'hi': k} != {'hi': k})
            print((k,) != (k,))

            print(k != l) # False as 5 != 6 + 1
            print(l != k) # True as 6 != 5 + 1
            print([l] != [k]) # True
            print({'hi': l} != {'hi': k}) # True
            print((l,) != (k,))
            """)

    def test_ne_reflected_within_datatypes(self):
        self.assertCodeExecution("""
            class F:
                def __init__(self, v):
                    self.v = v
                def __eq__(self, other):
                    return NotImplemented
            class G:
                def __init__(self, v):
                    self.v = v
                def __ne__(self, other):
                    return self.v != other.v + 1
            f = F(5)
            g = G(6)

            print([f] != [g]) # True (reflected)
            print([g] != [f]) # True (not reflected)
            print({'hi': f} != {'hi': g})
            print({'hi': g} != {'hi': f})
            print((f,) != (g,))
            print((g,) != (f,))
            """)

    def test_cmp_notimplemented(self):
        self.assertCodeExecution("""
            class A:
                pass
            x = A()

            # NotImplemented, not NotImplementedError
            print(x.__le__(x))
            print(x.__lt__(x))
            print(x.__ge__(x))
            print(x.__gt__(x))
            """)

    # next few tests from cpython's Lib/test/test_compare.py @ v3.6.0
    def test_comparisons(self):
        self.assertCodeExecution("""
            class Cmp:
                def __init__(self,arg):
                    self.arg = arg
                def __repr__(self):
                    return '<Cmp %s>' % self.arg
                def __eq__(self, other):
                    return self.arg == other
            class Empty:
                def __repr__(self):
                    return '<Empty>'

            set1 = [2, 2.0, 2+0j, Cmp(2.0)]
            set2 = [[1], (3,), None, Empty()]
            candidates = set1 + set2
            for a in candidates:
                for b in candidates:
                    print(a == b)
            """)

    # list object does not have insert() implemented yet
    @expectedFailure
    def test_id_comparisons(self):
        self.assertCodeExecution("""
            class Empty:
                def __repr__(self):
                    return '<Empty>'

            # Ensure default comparison compares id() of args
            L = []
            for i in range(10):
                L.insert(len(L)//2, Empty())
            for a in L:
                for b in L:
                    print(a == b, id(a) == id(b), 'a=%r, b=%r' % (a, b))
            """)

    def test_ne_defaults_to_not_eq(self):
        self.assertCodeExecution("""
            class Cmp:
                def __init__(self,arg):
                    self.arg = arg

                def __repr__(self):
                    return '<Cmp %s>' % self.arg

                def __eq__(self, other):
                    return self.arg == other

            a = Cmp(1)
            b = Cmp(1)
            c = Cmp(2)
            print(a == b) # True
            print(a != b) # False
            print(a != c) # True
            """)

    def test_ne_high_priority(self):
        self.assertCodeExecution("""
            # object.__ne__() should allow reflected __ne__() to be tried
            class Left:
                # Inherits object.__ne__()
                def __eq__(self, *args):
                    print('Left.__eq__')
                    return NotImplemented
            class Right:
                def __eq__(self, *args):
                    print('Right.__eq__')
                    return NotImplemented
                def __ne__(self, *args):
                    print('Right.__ne__')
                    return NotImplemented
            Left() != Right()
            """)

    def test_ne_low_priority(self):
        self.assertCodeExecution("""
            # object.__ne__() should not invoke reflected __eq__()
            class Base:
                # Inherits object.__ne__()
                def __eq__(self, *args):
                    print('Base.__eq__')
                    return NotImplemented
            class Derived(Base):  # Subclassing forces higher priority
                def __eq__(self, *args):
                    print('Derived.__eq__')
                    return NotImplemented
                def __ne__(self, *args):
                    print('Derived.__ne__')
                    return NotImplemented
            print(Base() != Derived())
            """)

    @expectedFailure
    # failing because class C not recreated on each loop
    def test_other_delegation(self):
        self.assertCodeExecution("""
            # No default delegation between operations except __ne__()
            ops = (
                ('__eq__', lambda a, b: a == b),
                ('__lt__', lambda a, b: a < b),
                ('__le__', lambda a, b: a <= b),
                ('__gt__', lambda a, b: a > b),
                ('__ge__', lambda a, b: a >= b),
            )
            for name, func in ops:
                def unexpected(*args):
                    print('Unexpected operator method called')
                class C:
                    __ne__ = unexpected
                for other, _ in ops:
                    if other != name:
                        setattr(C, other, unexpected)
                if name == '__eq__':
                    print(func(C(), object())) # False
                else:
                    try:
                        print(func(C(), object()))
                    except TypeError as err:
                        print(err)
            """)

    def test_issue_1393(self):
        self.assertCodeExecution("""
            class Anything:
                def __eq__(self, other):
                    return True

                def __ne__(self, other):
                    return False

            x = lambda: None
            print(x == Anything())
            print(Anything() == x)
            y = object()
            print(y == Anything())
            print(Anything() == y)
            """)

    def test_multiple_comparisons(self):
        self.assertCodeExecution("""
            print(1 == 1 == 1 == 1)
            print(1 > 2 > 3)
            print(1 == 1 == 2)
            print(1 == 1 != 2)
            print(1 == 1 < 2)
            print(3 == 3 < 2)
            print(1 < 2 < 3 > 2)
            print(0 == 0 == 0 == 0)
            print('abc' == '123' != '123')
            print('abc' == '123' == '123')
            print('abc' != '123' == '123')
            print(len('abcd') == 4 == len('1234'))
            x = 100
            print(x == (50 + 50) < 2000)
            print(x == (50 + 50) > 2000)
            """)

    def test_bad_contains(self):
        self.assertCodeExecution("""
            try:
                print(0 in 0)
            except TypeError as e:
                print(e)

            try:
                print(0 in True)
            except TypeError as e:
                print(e)

            class MyClass():
                value = "I am not iterable!"
            x = MyClass()

            try:
                print(0 in x)
            except TypeError as e:
                print(e)
        """)

    def test_bad_not_contains(self):
        self.assertCodeExecution("""
            try:
                print(0 not in 0)
            except TypeError as e:
                print(e)

            try:
                print(0 not in True)
            except TypeError as e:
                print(e)

            class MyClass():
                value = "I am not iterable!"
            x = MyClass()

            try:
                print(0 not in x)
            except TypeError as e:
                print(e)
        """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class SumTests(TranspileTestCase):
    def test_sum_list(self):
        self.assertCodeExecution("""
            print(sum([1, 2, 3, 4, 5, 6, 7]))
            print(sum([[1, 2], [3, 4], [5, 6]], []))
        """)

    def test_sum_tuple(self):
        self.assertCodeExecution("""
            print(sum((1, 2, 3, 4, 5, 6, 7)))
        """)

    def test_sum_iterator(self):
        self.assertCodeExecution("""
            i = iter([1, 2])
            print(sum(i))
            print(sum(i))
        """)

    def test_sum_mix_floats_and_ints(self):
        self.assertCodeExecution("""
            print(sum([1, 1.414, 2, 3.14159]))
        """)

    def test_sum_frozenset(self):
        self.assertCodeExecution("""
            print(sum(frozenset([1, 1.414, 2, 3.14159])))
        """)

    def test_sum_set(self):
        self.assertCodeExecution("""
            print(sum({1, 1.414, 2, 3.14159}))
        """)

    def test_sum_dict(self):
        self.assertCodeExecution("""
            print(sum({1: 1.414, 2: 3.14159}))
        """)

    def test_sum_generator_expressions(self):
        self.assertCodeExecution("""
            print(sum(x ** 2 for x in [3, 4]))
        """)


class BuiltinSumFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["sum"]

    not_implemented = [
    ]



from enum import Enum
from ..java import opcodes as JavaOpcodes


class ArgType(Enum):
    POSITIONAL_OR_KEYWORD = 1
    VAR_POSITIONAL = 2
    KEYWORD_ONLY = 3
    VAR_KEYWORD = 4


class OpcodePosition(Enum):
    START = 'start_op'
    END = 'end_op'
    NEXT = 'next_op'
    YIELD = 'yield_op'


##########################################################################
# Pseudo instructions used to flag the offset position of other
# attributes of the code, especially those depened on opcode offset.
##########################################################################

# A sentinel value to mark jumps that need to be resolved.
# This is used to differntiate from the "none" value indicating
# a currently open IF/TRY block.
RESOLVE = object()


class START_LOOP:
    """A pseudo instruction that can be used to mark the
    start of a loop (either a FOR or a WHILE)
    """
    def __init__(self):
        # The opcode that is the first in the formal LOOP block
        self.start_op = None

        # The last operation in the LOOP block
        self.end_op = None

        # If this IF can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Record the start of the if block
        context.next_resolve_list.append((self, OpcodePosition.START))

        # Record this loop.
        context.loops.append(self)

        # This opcode isn't for the final output.
        return False


class END_LOOP:
    """A no-op pseudo instruction that can be used to mark the
    end of a loop
    """
    def __init__(self):
        # The opcode that is the first in the formal IF block
        self.start_op = None

        # The last operation in the IF-ELSE block - the target for
        # all if-exiting jumps.
        self.end_op = None

        # If this IF can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Find the most recent if block on the stack that hasn't been
        # ended. That's the block that the elif applies to.
        for loop in context.loops[::-1]:
            if loop.end_op is None:
                loop.end_op = RESOLVE
                break

        # A jump back to the start of the loop for another iteration
        context.add_opcodes(jump(JavaOpcodes.GOTO(0), context, loop, OpcodePosition.START))

        # The next opcode is outside the LOOP block.
        context.next_resolve_list.append((loop, OpcodePosition.NEXT))
        context.next_resolve_list.append((loop, OpcodePosition.END))

        # The end operation is virtual, so all markers point
        # to the next operation
        context.next_resolve_list.append((self, OpcodePosition.START))
        context.next_resolve_list.append((self, OpcodePosition.END))
        context.next_resolve_list.append((self, OpcodePosition.NEXT))

        # This opcode isn't for the final output.
        return False


class IF:
    """Mark the start of an if-endif block.

    This forms the first of at least two pseudo-operations. It might be
    as simple as:
        IF([], IFOPCODE)
            ...
        END_IF()

    or as complex as

        IF([
                ... # commands to prepare stack for IF opcode
            ],
            IFOPCODE
        )
            ...
        ELIF([
                ... # commands to prepare stack for next IF opcode
            ],
            IFOPCODE
        )
            ...
        ELIF([
                ... # commands to prepare stack for next IF opcode
            ],
            IFOPCODE
        )
            ...
        ELSE()
            ...
        END_IF()

    END_IF comes *after* the last operation in the relevant block;
    but IF, ELIF and ELSE come *before* the first instruction. A little
    special handling is required to account for this.

        1. When the IF is created, it is empty, and has no start_op.
           The empty pseudo-instruction is added to the blocks list.
        2. When the instruction *after* the IF, ELIF or ELSE is processed,
           it is post-processed to fill in the operation marking the
           start of the relevant section.
        3. When the first ELIF/ELSE is found, that marks the end of the IF; a
           new elif block is registered. A GOTO is added, with no offset,
           to the end of the IF block. The offset will be updated
           when the END_IF is found.
        3. When the 2+ ELIF/ELSE is found, that marks the end of the previous
           ELIF; a new elif block is registered. A GOTO is added, with no offset,
           to the end of the previous block. The offset will be updated
           when the END_IF is found.
        4. When an END_IF is found, that marks the end of the current
           IF-ELIF-ELSE block. An end_op is recorded; that means this if block
           is no longer current.

    IF-ELSE blocks can be nested, so when an ELIF, ELSE or END_IF is found,
    it is recorded against the last IF in the blocks list
    that has no end_op recorded.

    """
    def __init__(self, commands, opcode):
        # The commands to prepare the stack for the IF comparison
        self.commands = commands if commands is not None else []
        # print("CREATE IF", id(self), opcode)

        # The opcode class to instantiate
        self.opcode = opcode

        # The opcode that is the first in the formal IF block
        self.start_op = None

        # The instantiated opcode for the comparison
        self.if_op = None

        # The list of all ELSE blocks associated with this IF
        self.elifs = []

        # The jump operation at the end of the IF, jumping to the
        # end of the IF-ELSE block.
        self.jump_op = None

        # The last operation in the IF-ELSE block - the target for
        # all if-exiting jumps.
        self.end_op = None

        # If this IF can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Record the start of the if block
        context.next_resolve_list.append((self, OpcodePosition.START))

        # Add the stack prepration commands to the code list
        context.add_opcodes(*self.commands)

        # Create an instance of the opcode and put it on the code list
        self.if_op = self.opcode(0)
        context.add_opcodes(self.if_op)

        # Record this IF.
        context.blocks.append(self)

        # This opcode isn't for the final output.
        return False


class ELIF:
    def __init__(self, commands, opcode):
        # The master IF block
        self.if_block = None

        # The commands to prepare the stack for the IF comparison
        self.commands = commands if commands is not None else []

        # The opcode class to instantiate
        self.opcode = opcode

        # The instantiated opcode for the comparison
        self.elif_op = None

        # The jump operation at the end of the ELIF, jumping to the
        # end of the IF-ELSE block.
        self.jump_op = None

        # If this ELIF can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Find the most recent if block on the stack that hasn't been
        # ended. That's the block that the elif applies to.
        for if_block in context.blocks[::-1]:
            if if_block.end_op is None:
                break

        # If this is the first elif, add a GOTO and use it as the
        # jump operation at the end of the IF block. If there are
        # ELIFs, add the GOTO as the jump operation on the most
        # recent ELIF.
        # However, if the most recent operation is a RETURN,
        # this jump isn't needed; just resolve the start position.
        if not isinstance(context.opcodes[-1], (JavaOpcodes.RETURN, JavaOpcodes.ARETURN)):
            jump_op = JavaOpcodes.GOTO(0)
            context.add_opcodes(jump_op)

            if len(if_block.elifs) == 0:
                if_block.jump_op = jump_op
            else:
                # print("    already got an endif")
                if_block.elifs[-1].jump_op = jump_op

        if len(if_block.elifs) == 0:
            jump(if_block.if_op, context, self, OpcodePosition.START)
        else:
            jump(if_block.elifs[-1].if_op, context, self, OpcodePosition.START)

        # Record the start of the elif block
        context.next_resolve_list.append((self, OpcodePosition.START))

        if self.opcode:
            # print("    this is an elif")
            # Add the stack prepration commands to the code list
            context.add_opcodes(*self.commands)

            # Create an instance of the opcode and put it on the code list
            self.if_op = self.opcode(0)
            context.add_opcodes(self.if_op)

        # else:
            # print("    this is an else")

        if_block.elifs.append(self)
        # print("IF BLOCKS: ", [(id(b), b.end_op) for b in context.blocks])

        # This opcode isn't for the final output.
        return False


class ELSE(ELIF):
    def __init__(self):
        # ELSE if an ELIF with no preparation and no comparison opcode
        super().__init__(None, None)


class END_IF:
    def __init__(self):
        # If this IF can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Find the most recent if block on the stack that hasn't been
        # ended. That's the block that the elif applies to.
        for if_block in context.blocks[::-1]:
            if if_block.end_op is None:
                if_block.end_op = RESOLVE
                break

        # If there aren't any ELIF/ELSE definitions, then the
        # main if block jumps straight to the end.
        if len(if_block.elifs) == 0:
            jump(if_block.if_op, context, if_block, OpcodePosition.NEXT)

        # Each of the 'end of block' jumps go to the end as well.
        if if_block.jump_op:
            jump(if_block.jump_op, context, if_block, OpcodePosition.NEXT)

        for block in if_block.elifs:
            if block.jump_op:
                jump(block.jump_op, context, if_block, OpcodePosition.NEXT)

        # The next opcode is outside the IF/ELIF/ELSE block.
        context.next_resolve_list.append((if_block, OpcodePosition.NEXT))

        # This opcode isn't for the final output.
        return False


class TRY:
    """Mark the start of a try-catch block.

    This forms the first of at least three pseudo-operations:
        TRY()
            ...
        CATCH(['your/exception/descriptor1', 'your/exception/descriptor2'])
            ...
        CATCH('your/other/exception')
            ...
        CATCH()
            ...
        FINALLY()
            ...
        END_TRY()

    END_TRY come *after* the last operation in the relevant block;
    but TRY and CATCH come *before* the first instruction. A little
    special handling is required to account for this.

        1. When the TRY is created, it is empty, and has no start_op.
           The empty pseudo-instruction is added to the exception list.
        2. When the instruction *after* the TRY or CATCH is processed,
           it is post-processed to fill in the operation marking the
           start of the relevant section.
        3. When a CATCH is found, that marks the end of the TRY; a
           new handler is registered in the current exception.
           If it's the first CATCH, a GOTO is added, with no offset,
           and recorded as the jump_op. The offset will be updated
           when the END is found.
        4. When an END is found, that marks the end of the current
           exception. An end_op is recorded; that means this try block
           is no longer current.

    Exception blocks can be nested, so when a CATCH or END is found,
    it is recorded against the last exception in the exception list
    that has no end_op recorded.

    """
    def __init__(self):
        # If this TRY can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

        # The first command covered by the try block
        self.start_op = None

        # The last command in the try block. This may be a
        # jump to the else or finally block, or past the
        # other handlers.
        self.try_end_op = None

        # The jump to the else/finally block.
        self.jump_op = None

        # The last command in the try/catch/else/finally sequence
        self.end_op = None

        # The catch handlers
        self.handlers = []

        # A handler for the finally block
        self.finally_handler = None

    def process(self, context):
        context.try_catches.append(self)
        # print("    try-catches", [(id(t), t.end_op) for t in context.try_catches])

        # The next opcode is the start of the TRY block.
        context.next_resolve_list.append((self, OpcodePosition.START))

        # This opcode isn't for the final output.
        return False


class CATCH:
    def __init__(self, descriptors=None):
        if descriptors is None:
            self.descriptors = []
        elif isinstance(descriptors, str):
            self.descriptors = [descriptors]
        else:
            self.descriptors = descriptors

        # If this CATCH can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

        # The first command covered by the catch block
        self.start_op = None

        # The end of the catch block.
        self.end_op = None

    # def __len__(self):
    #     # The CATCH needs to be able to pass as an opcode under initial
    #     # post processing
    #     return 3

    def process(self, context):
        # Find the most recent exception on the stack that hasn't been
        # ended. That's the block that the catch applies to.
        for try_catch in context.try_catches[::-1]:
            if try_catch.end_op is None:
                break

        # print("    current try_catch", try_catch)
        # print("    try-catches", [(id(t), t.end_op) for t in context.try_catches])

        # If this is the first catch, insert a GOTO operation to
        # the ELSE block. If it's the second or later catch, GOTO
        # the FINALLY block.
        # The jump distance will be updated when all the CATCH blocks
        # have been processed and the try_catch is converted.
        end_jump = JavaOpcodes.GOTO(0)
        if len(try_catch.handlers) == 0:
            try_catch.try_end_op = end_jump
            try_catch.jump_op = end_jump
        else:
            try_catch.handlers[-1].end_op = end_jump
            try_catch.handlers[-1].jump_op = end_jump
        context.add_opcodes(
            jump(end_jump, context, try_catch, OpcodePosition.NEXT)
        )

        # Add this catch block as a handler
        try_catch.handlers.append(self)

        # The next opcode is the start of the catch block.
        context.next_resolve_list.append((self, OpcodePosition.START))

        # This opcode isn't for the final output.
        return False


class FINALLY:
    def __init__(self):
        # If this FINALLY can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

        # The first command covered by the finally block
        self.start_op = None

        # The last command covered by the finally
        self.end_op = None

    def process(self, context):
        # Find the most recent exception on the stack that hasn't been
        # ended. That's the block that the catch applies to.
        for try_catch in context.try_catches[::-1]:
            if try_catch.end_op is None:
                break

        # print("    current try_catch", try_catch)
        # print("    try-catches", [(id(t), t.end_op) for t in context.try_catches])

        # If there are no catches, insert a GOTO operation.
        # The jump distance will be updated when all the CATCH blocks
        # have been processed and the try_catch is converted.
        # If it isn't the first catch, then this catch concludes the
        # previous one. Add a goto to the end of the block, and
        # record the end of the block for framing purposes.
        end_jump = JavaOpcodes.GOTO(0)
        if len(try_catch.handlers) == 0:
            try_catch.try_end_op = context.opcodes[-1]
        else:
            try_catch.handlers[-1].end_op = context.opcodes[-1]

        context.add_opcodes(jump(end_jump, context, try_catch, OpcodePosition.NEXT))

        # Add this block as the finally handler
        try_catch.finally_handler = self

        # The next opcode is the start of the finally block.
        context.next_resolve_list.append((self, OpcodePosition.START))

        # This opcode isn't for the final output.
        return False


class END_TRY:
    def __init__(self):
        # If this END TRY can be identified as the start of a new
        # line of source code, track that line.
        self.starts_line = None

    def process(self, context):
        # Find the most recent exception on the stack that hasn't been
        # ended. That's the block we're ending.
        for try_catch in context.try_catches[::-1]:
            if try_catch.end_op is None:
                try_catch.end_op = RESOLVE
                break

        # print("    current try_catch", try_catch)
        # print("    try-catches", [(id(t), t.end_op) for t in context.try_catches])

        if try_catch.finally_handler:
            try_catch.finally_handler.end_op = context.opcodes[-1]
        elif len(try_catch.handlers) > 0:
            try_catch.handlers[-1].end_op = context.opcodes[-1]

        try_catch.end_op = context.opcodes[-1]

        # The next opcode is the end of the try-catch block.
        context.next_resolve_list.append((try_catch, OpcodePosition.NEXT))

        # This opcode isn't for the final output.
        return False


##########################################################################
# Some opcodes need to reference other opcodes (e.g., GOTO)
#
# These helpers help define and resolve those references.
##########################################################################

def jump(opcode, context, target, position):
    """Define a jump operation.

    The specific offset will be resolved once all the
    Java opcodes have been instantiated
    """
    # print("    add jump to reference %s 0x%x %s %s..." % (opcode, id(opcode), target, position))
    context.unknown_jump_targets.setdefault(target, []).append((opcode, position))
    return opcode


def resolve_jump(opcode, context, target, position):
    """Resolve a jump target in an opcode.

    target is the Python AST node.
    When Python code is converted to Java, it will turn into
    0-N opcodes. We need to specify which one will be used
    as the Java offset:
     * START - the first Java opcode generated from this Python opcode
     * END - the last Java opcode generated from the Python opcode
     * NEXT - the next Java opcode added after this block.
    """
    # print("RESOLVE %s 0x%x to %s %s" % (opcode, id(opcode), target, position))
    if position == OpcodePosition.START:
        opcode.jump_op = target.start_op
    elif position == OpcodePosition.END:
        opcode.jump_op = target.end_op
    elif position == OpcodePosition.NEXT:
        opcode.jump_op = target.next_op
    elif position == OpcodePosition.YIELD:
        opcode.jump_op = target.yield_op
    else:
        raise Exception("Unknown opcode position")
    # print("JUMP OP IS", opcode.jump_op)
    context.jumps.append(opcode)
    opcode.jump_op.references.append(opcode)


##########################################################################
# Not entirely common, but complex, calls.
##########################################################################

class AddToArgs:
    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.INVOKESTATIC(
                'org/Python',
                'addToArgs',
                args=['[Lorg/python/Object;', 'Lorg/python/Object;'],
                returns='[Lorg/python/Object;'
            )
        )


class AddToKwargs:
    def __init__(self, kwarg):
        self.kwarg = kwarg

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.LDC_W(self.kwarg),
            JavaOpcodes.INVOKESTATIC(
                'org/Python',
                'addToKwargs',
                args=['Ljava/util/Map;', 'Lorg/python/Object;', 'Ljava/lang/String;'],
                returns='Ljava/util/Map;'
            ),
        )



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class PrintTests(TranspileTestCase):
    def test_fileobj(self):
        self.assertCodeExecution("""
            class FileLikeObject:
                def __init__(self):
                    self.buffer = ''

                def write(self, content):
                    self.buffer = self.buffer + (content * 2)

            out = FileLikeObject()

            print('hello', 'world', file=out)
            print('goodbye', 'world', file=out)
            print()
            """)

    def test_sep(self):
        self.assertCodeExecution("""
            print('hello world', 'goodbye world', sep='-')
            print()
            """)

    def test_end(self):
        self.assertCodeExecution("""
            print('hello world', 'goodbye world', end='-')
            print()
            """)

    def test_flush(self):
        self.assertCodeExecution("""
            print('hello world', 'goodbye world', flush=True)
            print()
            """)

    def test_combined(self):
        self.assertCodeExecution("""
            class FileLikeObject:
                def __init__(self):
                    self.buffer = ''

                def write(self, content):
                    self.buffer = self.buffer + (content * 2)

                def flush(self):
                    self.buffer = self.buffer + '<<<'

            out = FileLikeObject()

            print('hello', 'world', sep='*', end='-', file=out, flush=True)
            print('goodbye', 'world', file=out, sep='-', end='*')
            print()
            """)


class BuiltinPrintFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["print"]

    not_implemented = [
        'test_class',
    ]



from unittest import expectedFailure

from ..utils import TranspileTestCase


class MethodTests(TranspileTestCase):
    def test_method(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, value):
                    print(value * 3)
                    return value + 5

            obj = TestObj()
            print("value =", obj.myfunc(5))
            """)

    def test_noargs_method(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, ):
                    print('Hello')
                    return 5

            obj = TestObj()
            print("value =", obj.myfunc())
            """)

    def test_void_method(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, value):
                    print(value * 3)

            obj = TestObj()
            obj.myfunc(5)
            """)

    def test_inner_function(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, value):
                    print(value * 3)

                    def myinner(value2):
                        print(value2 * 4)
                        return value2 + 6

                    print("inner value =", myinner(10))
                    return value + 5

            obj = TestObj()
            print("outer =", obj.myfunc(5))
            """, run_in_function=False)

    def test_closure(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, value):
                    print(value * 3)

                    def myinner(value2):
                        print(value2 * value)
                        return value2 + 6

                    print("inner value =", myinner(10))
                    return value + 5

            obj = TestObj()
            print("outer =", obj.myfunc(5))
            """)

    def test_default_args(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(37)
            """, run_in_function=False)

    def test_override_some_default_args(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(37, 42)
            """, run_in_function=False)

    def test_overide_all_default_args(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(37, 42, 99)
            """, run_in_function=False)

    def test_use_kwargs(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(37, y=42)
            """, run_in_function=False)

    def test_use_kwargs_non_sequential(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(37, z=42)
            """, run_in_function=False)

    def test_use_all_kwargs(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(x=37, y=42, z=99)
            """, run_in_function=False)

    def test_use_all_kwargs_different_order(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, x, y=1, z=2):
                    print("x =", x)
                    print("y =", y)
                    print("z =", z)

            obj = TestObj()
            obj.myfunc(z=99, y=42, x=37)
            """, run_in_function=False)

    def test_call_method_with_var_args(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, *args):
                    print(args)
                    return args[0]

            obj = TestObj()
            print("first arg =", obj.myfunc(1, 2, 3, 4, 5))
            """, run_in_function=False)

    def test_call_method_with_kw(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, **kwargs):
                    print(kwargs['first'] * 3)
                    print(kwargs['second'] * 3)

                    return kwargs['first'] + kwargs['second']

            obj = TestObj()
            print("values sum =", obj.myfunc(first=1, second=2))
            """, run_in_function=False)

    def test_call_method_kw(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, **kwargs):
                    print(kwargs['first'] * 3)
                    print(kwargs['second'] * 3)

                    return kwargs['first'] + kwargs['second']

            obj = TestObj()
            values = {'first': 1, 'second': 2}
            print("values sum =", obj.myfunc(**values))
            """, run_in_function=False)

    def test_call_method_var_kw(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, *args, **kwargs):
                    print(kwargs['first'] * 3)
                    print(kwargs['second'] * 3)
                    print(args)

                    return kwargs['first'] + kwargs['second']

            obj = TestObj()
            values_tuple = (1, 2, 3, 4)
            values_dict = {'first': 1, 'second': 2}
            print("values sum =", obj.myfunc(*values_tuple, **values_dict))
            """, run_in_function=False)

    def test_call_method_var(self):
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self, *args):
                    print(args)

                    return len(args)

            obj = TestObj()
            values_tuple = (1, 2, 3, 4)
            print("values count =", obj.myfunc(*values_tuple))
            """, run_in_function=False)

    @expectedFailure
    def test_method_caching_not_visible_in_dict(self):
        # Method caching is a performance optimization that should not affect __dict__ inspection
        self.assertCodeExecution("""
            class TestObj:
                def myfunc(self):
                    print(0)

            obj = TestObj()
            obj.myfunc()

            try:
                print(obj.__dict__)
            except AttributeError as err:
                print(err)
        """, run_in_function=False)




from ..utils import TranspileTestCase


class WhileLoopTests(TranspileTestCase):
    def test_while(self):
        self.assertCodeExecution("""
            i = 0
            total = 0
            while i < 10:
                i += 1
                total += i
                print(i, total)
            """)

        # Test expression statement
        self.assertCodeExecution("""
            x = True
            while x:
                [1, 2, 3]
                "abc" == "def"
                x = False
            """)

    def test_break(self):
        self.assertCodeExecution("""
            i = 0
            while i < 10:
                i = i + 1
                print(i, i % 5)
                if i % 5 == 0:
                    break
                print("after")
            """)

    def test_continue(self):
        self.assertCodeExecution("""
            i = 0
            while i < 10:
                i = i + 1
                print(i, i % 5)
                if i % 5 == 0:
                    continue
                print("after")
            """)

    def test_nested(self):
        self.assertCodeExecution("""
            i = 1
            j = 10
            while i < j:
                k = 0
                while k < i:
                    print(i, j)
                    k = k + 1
                print("While done")
                i = i + 1
            """)

    def test_while_forever(self):
        self.assertCodeExecution("""
            i = 0
            while 1:
                print("Loop", i)
                i = i + 1
                if i == 10:
                    break
            """)

    def test_while_forever_with_if_not(self):
        self.assertCodeExecution("""
            i = 0
            while 1:
                print("Loop", i)
                i = i + 1
                if not i < 10:
                    break
            """)

    def test_while_not_forever(self):
        self.assertCodeExecution("""
            while not 0:
                print("Loop")
                break
            """)

    def test_while_else(self):
        self.assertCodeExecution("""
            i = 1
            j = 4
            while i < j:
                print(i)
                i = i + 1
            else:
                print("Else")
            """)

    def test_while_else_break(self):
        self.assertCodeExecution("""
            i = 1
            j = 4
            while i < j:
                print(i)
                i = i + 1
                break
            else:
                print("Else")
            """)

    def test_while_forever_inside_try(self):
        self.assertCodeExecution("""
            i = 0
            try:
                while 1:
                    print("Loop", i)
                    i = i + 1
                    if i == 10:
                        break
            finally:
                print("Done")
            """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class ZipTests(TranspileTestCase):
    def test_zip(self):
        self.assertCodeExecution("""
            lst=['a','b','c','d','e']
            str="TestString"
            tup=(1,2,5,9,4,8,0,4)
            print(list(zip(lst,str,tup)))
            print(list(zip(lst,tup,"hi there")))
            """)

    def test_zip_expectedFailure(self):
        self.assertCodeExecution("""
            lst=['a','b','c','d','e']
            str="TestString"
            tup=(1,2,5,9,4,8,0,4)
            try:
                print(list(zip(5)))
            except TypeError as err:
                print(err)
            try:
                print(list(zip(lst,str,tup,7.9)))
            except TypeError as err:
                print(err)
            try:
                print(list(zip(True,lst,tup,"hi there")))
            except TypeError as err:
                print(err)
            """)


class BuiltinZipFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["zip"]

    not_implemented = [
        'test_bytearray',
        'test_bytes',
        'test_dict',
        'test_frozenset',
        'test_list',
        'test_range',
        'test_set',
        'test_str',
        'test_tuple',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class AnyTests(TranspileTestCase):
    pass


class BuiltinAnyFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["any"]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class BytearrayTests(TranspileTestCase):
    pass


class BuiltinBytearrayFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["bytearray"]



from .. utils import (
    TranspileTestCase,
    UnaryOperationTestCase,
    BinaryOperationTestCase,
    InplaceOperationTestCase,
    SAMPLE_SUBSTITUTIONS
)


class FloatTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = 3.14159
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = 3.14159
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = 3.14159
            try:
                x[0] = 2
            except TypeError as err:
                print(err)
        """)

    def test_repr(self):
        self.assertCodeExecution("""
            x = 350000000000000000.0
            print(x)
            x = 3500.0
            print(x)
            x = 35.0
            print(x)
            x = 3.5
            print(x)
            x = 0.35
            print(x)
            x = 0.035
            print(x)
            x = 0.0035
            print(x)
            x = 0.00035
            print(x)
            x = 0.000035
            print(x)
            x = 0.0000035
            print(x)
            x = 0.00000000000000035
            print(x)

            x = 0.0
            print(x)
            x = float('-0.0')
            print(x)
            x = float('nan')
            print(x)
            x = float('inf')
            print(x)
            x = float('-inf')
            print(x)
            """)

    def test_negative_zero_constant(self):
        self.assertCodeExecution("""
            x = -0.0
            y = 0.0
            print(x, y)
            """)

    def test_is_integer(self):
        self.assertCodeExecution("""
            x = 0.0
            print(x.is_integer())
            x = 3.14
            print(x.is_integer())
            x = -1.0
            print(x.is_integer())
            x = -62.5
            print(x.is_integer())
            x = float('nan')
            print(x.is_integer())
            x = float('inf')
            print(x.is_integer())
            x = float('-inf')
            print(x.is_integer())
            """)

    def test_hex(self):
        numbers = [
            0e0, -0e0, 10000152587890625e-16, -566e85,
            -87336362425182547697e-280, 4.9406564584124654e-324,
            'nan', 'inf', '-inf'
        ]
        template = """
            x = float('{}')
            print(x.hex())
            """
        code = '\n'.join(template.format(number) for number in numbers)
        self.assertCodeExecution(code)

    def test_mul_TypeError(self):
        self.assertCodeExecution("""
            a = 5.6
            try:
                print(a*None);
            except TypeError as e:
                print(e)
            """)

    def test_none(self):
        self.assertCodeExecution("""
            try:
                print(float(None))
            except TypeError as err:
                print(err)
        """)

    def test_no_arguments(self):
        self.assertCodeExecution("""
            print(float())
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(float(1, 2))
            except TypeError as err:
                print(err)
        """)


class UnaryFloatOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'float'

    not_implemented = [
    ]


class BinaryFloatOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'float'

    substitutions = {
        "(-0.8946025309573877+0.446862743585361j)": [
            "(-0.8946025309573877+0.44686274358536093j)"
        ]
    }

    substitutions.update(SAMPLE_SUBSTITUTIONS)

    not_implemented = [
    ]


class InplaceFloatOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'float'

    substitutions = {
        "(-0.8946025309573877+0.446862743585361j)": [
            "(-0.8946025309573877+0.44686274358536093j)"
        ]
    }

    substitutions.update(SAMPLE_SUBSTITUTIONS)

    not_implemented = [
    ]



from ..utils import TranspileTestCase


class DocstringTests(TranspileTestCase):
    def test_method_docstring(self):
        self.assertCodeExecution("""
            def test():
                "This is the docstring"
                return 3

            print(test())
            print(test.__doc__)
            """)

    def test_naked_string(self):
        self.assertCodeExecution("""
            def test():
                x = 3
                "This is a naked string"
                return x

            print(test())
            print(test.__doc__)
            """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class FloatTests(TranspileTestCase):
    pass


class BuiltinFloatFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["float"]

    not_implemented = [
        'test_bytearray',
        'test_bytes',
        'test_complex',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase

from unittest import expectedFailure


class AbsTests(TranspileTestCase):
    def test_abs_not_implemented(self):
        self.assertCodeExecution("""
            class NotAbsLike:
                pass
            x = NotAbsLike()
            try:
                print(abs(x))
            except TypeError as err:
                print(err)
            """)

    @expectedFailure
    def test_incorrect_abs_call(self):
        self.assertCodeExecution("""
            x = 1
            print(x.abs())
            """)


class BuiltinAbsFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["abs"]

    not_implemented = [
    ]



from .. utils import TranspileTestCase, BuiltinTwoargFunctionTestCase


class GetattrTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            class MyClass(object):
                class_value = 42

                def __init__(self, val):
                    self.value = val

                def stuff(self, delta):
                    print("DELTA: ", delta)
                    return self.value + delta

            print("On class: ")
            print('  class_value =', getattr(MyClass, 'class_value'))
            # print('  stuff =', getattr(MyClass, 'stuff'))  # FIXME
            try:
                getattr(MyClass, 'foo')
                print("  Shouldn't be able to get attribute foo")
            except AttributeError:
                print("  Can't get attribute foo")
            print('  foo (default) =', getattr(MyClass, 'foo', 42))

            obj = MyClass(37)

            print("On instance:")
            print('  class_value =', getattr(obj, 'class_value'))
            print('  value =', getattr(obj, 'value'))
            # print('  stuff =', getattr(obj, 'stuff'))  # FIXME
            try:
                getattr(MyClass, 'foo')
                print("  Shouldn't be able to get attribute foo")
            except AttributeError:
                print("  Can't get attribute foo")
            print('  foo (default) =', getattr(obj, 'foo', 42))
            """, run_in_function=False)


class BuiltinGetattrFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["getattr"]

    not_implemented = [
        'test_class_str',
        'test_obj_str',
    ]



import types

from ..java import (
    Code as JavaCode, ExceptionInfo as JavaExceptionInfo, LineNumberTable,
    opcodes as JavaOpcodes, Classref
)
from .structures import (
    ArgType, OpcodePosition,
    TRY, CATCH, END_TRY,
    jump, resolve_jump,
)
from .types import java, python
from .types.primitives import (
    ICONST_val, LCONST_val, DCONST_val, ALOAD_index, ALOAD_name, ASTORE_index, ASTORE_name, free_name
)

# from .debug import DEBUG, DEBUG_value


class IgnoreBlock(Exception):
    """An escape hatch; enable a block to be flagged as ignorable"""
    pass


class BlockCodeTooLarge(Exception):
    """Enable a block to be flagged too large to transpile."""
    def __init__(self, code_length):
        self.code_length = code_length


class Accumulator:
    def __init__(self, local_vars=None):
        self.opcodes = []

        self.local_vars = local_vars if local_vars else {}
        self.deleted_vars = set()

        self.next_resolve_list = []
        self.next_opcode_starts_line = None

    def add_opcodes(self, *opcodes):
        # Add the opcodes to the code list and process them.
        for opcode in opcodes:
            if opcode.process(self):
                # self.opcodes.extend([
                #     DEBUG(str(opcode)),
                # ])

                self.opcodes.append(opcode)

                # If we've flagged a code line change, attach that to the opcode
                if self.next_opcode_starts_line:
                    opcode.starts_line = self.next_opcode_starts_line
                    self.next_opcode_starts_line = None

                # Resolve any references to the "next" opcode.
                for (obj, attr) in self.next_resolve_list:
                    # print("        resolve %s reference on %s %s with %s %s" % (
                    #     attr, obj, id(obj), opcode, id(opcode))
                    # )
                    setattr(obj, attr.value, opcode)

                self.next_resolve_list = []

    @property
    def active_local_vars(self):
        return {
            name: index
            for name, index in self.local_vars.items()
            if index is not None
        }

    def stack_depth(self):
        "Evaluate the maximum stack depth required by a sequence of Java opcodes"
        depth = 0
        max_depth = 0

        for opcode in self.opcodes:
            depth = depth + opcode.stack_effect
            # print("   ", opcode, depth)
            if depth > max_depth:
                max_depth = depth
        return max_depth

    def max_stack(self, exceptions=None):
        stack_depth = self.stack_depth()
        if exceptions:
            return stack_depth + len(exceptions)
        else:
            return stack_depth

    def max_locals(self):
        return len(self.active_local_vars) + len(self.deleted_vars)


class Block(Accumulator):
    def __init__(self, parent=None, verbosity=0):
        super().__init__()
        self._parent = parent
        self.verbosity = verbosity

        self.has_self = False
        self.parameters = []
        self.symbols = {}

        self.generator = None
        self.yield_points = []

        self.try_catches = []
        self.blocks = []
        self.jumps = []
        self.loops = []
        self.unknown_jump_targets = {}
        self.returns = {
            'annotation': None
        }

    def store_name(self, name, declare=False):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def store_dynamic(self):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def load_name(self, name):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def load_globals(self):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def load_locals(self):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def load_vars(self):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    def delete_name(self, name):
        raise NotImplementedError('Abstract class `block` cannot be used directly.')

    @property
    def can_ignore_empty(self):
        return False

    @property
    def has_nested_structure(self):
        """Whether the block has nested structures inside.
        """
        return any([self.blocks, self.loops, self.try_catches])

    def add_str(self, value):
        self.add_opcodes(
            python.Str(value),
        )

    def add_int(self, value):
        self.add_opcodes(
            LCONST_val(value),
            JavaOpcodes.INVOKESTATIC('org/python/types/Int', 'getInt', args=['J'], returns='Lorg/python/types/Int;'),
        )

    def add_float(self, value):
        self.add_opcodes(
            java.New('org/python/types/Float'),
            DCONST_val(value),
            java.Init('org/python/types/Float', 'D'),
        )

    def add_complex(self, value):
        self.add_opcodes(
            java.New('org/python/types/Complex'),
            DCONST_val(value.imag),
            java.Init('org/python/types/Complex', 'D'),
        )

    def _add_value(self, val):
        if isinstance(val, bool):
            if val is True:
                self.add_opcodes(
                    JavaOpcodes.GETSTATIC('org/python/types/Bool', 'TRUE', 'Lorg/python/types/Bool;'),
                )
            else:
                self.add_opcodes(
                    JavaOpcodes.GETSTATIC('org/python/types/Bool', 'FALSE', 'Lorg/python/types/Bool;'),
                )

        elif isinstance(val, int):
            self.add_int(val)

        elif isinstance(val, float):
            self.add_opcodes(
                java.New('org/python/types/Float'),
                JavaOpcodes.LDC2_W(val),
                java.Init('org/python/types/Float', 'D'),
            )

        elif isinstance(val, str):
            self.add_opcodes(
                python.Str(val),
            )

        elif isinstance(val, bytes):
            self.add_opcodes(
                java.New('org/python/types/Bytes'),
                JavaOpcodes.LDC_W(val.decode('ISO-8859-1')),
                java.Init('org/python/types/Bytes', 'Ljava/lang/String;'),
            )

        elif isinstance(val, tuple):
            self.add_tuple(val)

        elif isinstance(val, complex):
            self.add_opcodes(
                java.New('org/python/types/Complex'),
                DCONST_val(val.real),
                DCONST_val(val.imag),
                java.Init('org/python/types/Complex', 'D', 'D'),
            )

        elif isinstance(val, types.CodeType):
            self.add_opcodes(
                JavaOpcodes.ACONST_NULL()
            )

        else:
            raise RuntimeError("Unknown constant type %s" % type(val))

    def add_list(self, data):
        self.add_opcodes(
            java.New('java/util/ArrayList'),
            java.Init('java/util/ArrayList'),
        )

        for value in data:
            self.add_opcodes(
                JavaOpcodes.DUP(),
            )

            if value is None:
                self.add_opcodes(
                    python.NONE()
                )
            elif isinstance(value, frozenset):
                self.add_opcodes(
                    java.New('org/python/types/FrozenSet'),
                    python.Set(),
                )
                for elt in value:
                    self.add_opcodes(
                        JavaOpcodes.DUP()
                    )

                    self._add_value(elt)

                    self.add_opcodes(
                        python.Set.add()
                    )
                self.add_opcodes(
                    java.Init(
                        'org/python/types/FrozenSet',
                        'Lorg/python/types/Set;',
                    )
                )
            else:
                self._add_value(value)

            self.add_opcodes(
                java.List.add(),
            )

    def add_tuple(self, data):
        self.add_opcodes(
            java.New('org/python/types/Tuple'),
        )
        self.add_list(data)
        self.add_opcodes(
            java.Init('org/python/types/Tuple', 'Ljava/util/List;'),
        )

    def add_callable(self, function, closure=False):
        self.add_opcodes(
            TRY(),
        )
        self.add_opcodes(
                # Wrap that function into a Callable.
                java.New('org/python/types/Function'),
        )

        self.add_str(function.code.co_name)
        self.add_opcodes(
            JavaOpcodes.LDC_W(function.code.co_argcount),
            JavaOpcodes.LDC_W(function.code.co_kwonlyargcount),
            JavaOpcodes.LDC_W(function.code.co_flags),
        )
        self.add_tuple(function.code.co_consts)
        self.add_list(function.code.co_varnames)

        self.add_opcodes(
                # Get a Java Method representing the new function
                JavaOpcodes.LDC_W(Classref(function.class_descriptor)),
                JavaOpcodes.LDC_W(function.pyimpl_name),

                java.Array(len(function.parameters), classname='java/lang/Class'),
        )

        for i, param in enumerate(function.parameters):
            self.add_opcodes(
                JavaOpcodes.DUP(),
                ICONST_val(i),
                JavaOpcodes.LDC_W(Classref('org/python/Object')),
                JavaOpcodes.AASTORE(),
            )

        self.add_opcodes(
                JavaOpcodes.INVOKEVIRTUAL(
                    'java/lang/Class',
                    'getMethod',
                    '(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;'
                ),

                # globals
                # JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),
                JavaOpcodes.ACONST_NULL(),  # globals

                # Default args
                java.List(),
        )

        # Default arguments list
        for arg in function.parameters:
            if arg['kind'] == ArgType.POSITIONAL_OR_KEYWORD and arg['default']:
                self.add_opcodes(
                    JavaOpcodes.DUP(),
                    ALOAD_name(arg['default']),
                    java.List.add(),
                )

        # Default keyword arguments list
        self.add_opcodes(
                java.Map(),
        )

        for arg in function.parameters:
            if arg['kind'] in (
                    ArgType.POSITIONAL_OR_KEYWORD,
                    ArgType.KEYWORD_ONLY) and arg['default']:
                self.add_opcodes(
                    JavaOpcodes.DUP(),
                    JavaOpcodes.LDC_W(arg['name']),
                    ALOAD_name(arg['default']),
                    java.Map.put(),
                )

        # Closure. This is created and assigned when the function is accessed.
        self.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
        )

        self.add_opcodes(
                java.Init(
                    'org/python/types/Function',
                    'Lorg/python/types/Str;',
                    'I',
                    'I',
                    'I',
                    'Lorg/python/types/Tuple;',
                    'Ljava/util/List;',
                    'Ljava/lang/reflect/Method;',
                    'Ljava/util/Map;',
                    'Ljava/util/List;',
                    'Ljava/util/Map;',
                    'Lorg/python/types/Closure;',
                ),
        )
        self.add_opcodes(
            CATCH('java/lang/NoSuchMethodError')
        )
        self.add_opcodes(
                ASTORE_name('#EXCEPTION#'),
                java.New('org/python/exceptions/RuntimeError'),
                JavaOpcodes.LDC_W('Unable to find MAKE_FUNCTION output %s.%s' % (
                    function.class_descriptor, function.name)
                ),
                java.Init('org/python/exceptions/RuntimeError', 'Ljava/lang/String;'),
                JavaOpcodes.ATHROW(),
        )
        self.add_opcodes(
            END_TRY(),
            free_name('#EXCEPTION#')
        )

    def visitor_setup(self):
        """Tweak the bytecode generated for this block."""
        pass

    def visitor_teardown(self):
        """Tweak the bytecode generated for this block."""
        pass

    def transpile_code(self):
        """Create a JavaCode object representing the opcodes stored in the block

        May raise ``IgnoreBlock`` if the block should be ignored.
        """
        # Install the shortcut jump points for yield statements.
        yield_jumps = Accumulator()

        for i, yield_point in enumerate(self.yield_points):
            yield_jumps.add_opcodes(
                ALOAD_index(self.local_vars['<generator>']),
                JavaOpcodes.GETFIELD('org/python/types/Generator', 'yield_point', 'I'),
                ICONST_val(i + 1),
                jump(JavaOpcodes.IF_ICMPEQ(0), self, yield_point, OpcodePosition.YIELD)
            )

        self.opcodes = yield_jumps.opcodes + self.opcodes

        # Make sure every local variable slot has been initialized
        # as an object. This is needed because Python allows a variable
        # to be instantiated in a sub-block, and used outside that block.
        # The JVM doesn't, and raises a verify error if you try. By
        # initializing all variables, we can trick the verifier.
        # TODO: Ideally, we'd only initialize the variables that are ambiguous.
        init_vars = Accumulator()
        for i in range(
                    len(self.parameters) + (1 if self.has_self else 0),
                    len(self.active_local_vars) + len(self.deleted_vars)
                ):
            init_vars.add_opcodes(
                JavaOpcodes.ACONST_NULL(),
                ASTORE_index(i)
            )

        self.opcodes = init_vars.opcodes + self.opcodes

        # Since we've processed all the Python opcodes, we can now resolve
        # all the unknown jump targets.
        # print('>>>>> Resolve references')
        for target, references in self.unknown_jump_targets.items():
            # print("   resolving %s references to %s" % (len(references), target))
            for opcode, position in references:
                resolve_jump(opcode, self, target, position)

        # If the block has no content in it, and the block allows,
        # ignore this block.
        if self.can_ignore_empty:
            if len(self.opcodes) == 1 and isinstance(self.opcodes[0], JavaOpcodes.RETURN):
                raise IgnoreBlock()
            elif len(self.opcodes) == 2 and isinstance(self.opcodes[1], JavaOpcodes.ARETURN):
                raise IgnoreBlock()

        # Now that we have a complete opcode list, postprocess the list
        # with the known offsets.
        offset = 0
        # print('>>>>> set offsets', self)
        for index, instruction in enumerate(self.opcodes):
            # print("%4d:%4d (0x%x) %s" % (index, offset, id(instruction), instruction))
            instruction.java_index = index
            instruction.java_offset = offset
            offset += len(instruction)
        # print('>>>>> end set offsets')

        # The maximum length of a method in JVM is 65534 bytes. (See: Section 4.7.3, JVM 7 Specs)
        if offset > 65534:
            raise BlockCodeTooLarge(offset)

        # Construct the exception table, updating any
        # end-of-exception GOTO operations with the right opcode.
        # Record a frame range for each one.
        exceptions = []
        for try_catch in self.try_catches:
            # print("TRY CATCH START", id(try_catch), try_catch.start_op, try_catch.start_op.java_offset)
            # print("        TRY END", try_catch.try_end_op, try_catch.try_end_op.java_offset)
            # print("            END", try_catch.end_op, try_catch.end_op.java_offset)
            for handler in try_catch.handlers:
                # print("  HANDLER", handler.start_op, handler.end_op, handler.descriptors)
                if handler.descriptors:
                    for descriptor in handler.descriptors:
                        exceptions.append(JavaExceptionInfo(
                            try_catch.start_op.java_offset,
                            try_catch.try_end_op.java_offset,
                            handler.start_op.java_offset,
                            descriptor
                        ))
                else:
                    exceptions.append(JavaExceptionInfo(
                        try_catch.start_op.java_offset,
                        try_catch.try_end_op.java_offset,
                        handler.start_op.java_offset,
                        'org/python/exceptions/BaseException'
                    ))

            # Add definitions for the finally block
            if try_catch.finally_handler:
                # print(
                #     "  FINALLY",
                #     try_catch.finally_handler.start_op.java_offset,
                #     try_catch.finally_handler.end_op.java_offset
                # )
                exceptions.append(JavaExceptionInfo(
                    try_catch.start_op.java_offset,
                    try_catch.try_end_op.java_offset,
                    try_catch.finally_handler.start_op.java_offset,
                    None
                ))
                for handler in try_catch.handlers:
                    # print("   h", handler.descriptors)
                    exceptions.append(JavaExceptionInfo(
                        handler.start_op.java_offset,
                        handler.end_op.java_offset,
                        try_catch.finally_handler.start_op.java_offset,
                        None
                    ))

        # Update any jump instructions
        # print ("There are %s jumps" % len(self.jumps))
        for jmp in self.jumps:
            # print ("JUMP", hex(id(jmp)), jmp, jmp.java_offset, jmp.jump_op, hex(id(jmp.jump_op)))
            try:
                jmp.offset = jmp.jump_op.java_offset - jmp.java_offset
            except AttributeError:
                jmp.offset = jmp.jump_op.start_op.java_offset - jmp.java_offset

        # Construct a line number table from
        # the source code reference data on opcodes.
        line_numbers = []
        for opcode in self.opcodes:
            if opcode.starts_line is not None:
                line_numbers.append((opcode.java_offset, opcode.starts_line))
        line_number_table = LineNumberTable(line_numbers)

        return JavaCode(
            max_stack=self.max_stack(exceptions),
            max_locals=self.max_locals(),
            code=self.opcodes,
            exceptions=exceptions,
            attributes=[
                line_number_table
            ]
        )

    def transpile(self):
        return self.transpile_code()



import unittest

from ..utils import TranspileTestCase


class SequenceTests(TranspileTestCase):
    def test_unpack_sequence(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            a, b, c = x
            print(a)
            print(b)
            print(c)
            """)

    @unittest.expectedFailure
    def test_unpack_sequence_overflow(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            a, b = x
            print(a)
            print(b)
            """)

    @unittest.expectedFailure
    def test_unpack_sequence_underflow(self):
        self.assertCodeExecution("""
            x = [1, 2]
            a, b, c = x
            print(a)
            print(b)
            print(c)
            """)



import sys

from unittest import expectedFailure

from ..utils import TranspileTestCase


class SysModuleTests(TranspileTestCase):

    ############################################################
    # __displayhook__
    @expectedFailure
    def test___displayhook__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__displayhook__())
            """)

    ############################################################
    # __doc__
    def test___doc__(self):
        if sys.hexversion < 0x03040500:
            # A minor change in the docstring for 3.4.5
            substitutions = {
                'the value of the largest Unicode code point': [
                    'the value of the largest Unicode codepoint'
                ]
            }
        else:
            substitutions = None

        self.assertCodeExecution(
            """
            import sys
            print(sys.__doc__)
            """,
            substitutions=substitutions
        )

    ############################################################
    # __egginsert
    @expectedFailure
    def test___egginsert(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__egginsert)
            """)

    ############################################################
    # __excepthook__
    @expectedFailure
    def test___excepthook__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__excepthook__())
            """)

    ############################################################
    # __loader__
        self.assertCodeExecution("""
            import sys
            print(sys.__excepthook__())
            """)

    ############################################################
    # __name__
    def test___name__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__name__)
            """)

    ############################################################
    # __package__
    def test___package__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__package__)
            """)

    ############################################################
    # __plen
    @expectedFailure
    def test___plen(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__plen)
            """)

    ############################################################
    # __spec__
    @expectedFailure
    def test___spec__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__spec__())
            """)

    ############################################################
    # __stderr__
    @expectedFailure
    def test___stderr__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__stderr__())
            """)

    ############################################################
    # __stdin__
    @expectedFailure
    def test___stdin__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__stdin__())
            """)

    ############################################################
    # __stdout__
    @expectedFailure
    def test___stdout__(self):
        self.assertCodeExecution("""
            import sys
            print(sys.__stdout__())
            """)

    ############################################################
    # _clear_type_cache
    @expectedFailure
    def test__clear_type_cache(self):
        self.assertCodeExecution("""
            import sys
            print(sys._clear_type_cache())
            """)

    ############################################################
    # _current_frames
    @expectedFailure
    def test__current_frames(self):
        self.assertCodeExecution("""
            import sys
            print(sys._current_frames())
            """)

    ############################################################
    # _debugmallocstats
    @expectedFailure
    def test__debugmallocstats(self):
        self.assertCodeExecution("""
            import sys
            print(sys._debugmallocstats())
            """)

    ############################################################
    # getframe
    @expectedFailure
    def test__getframe(self):
        self.assertCodeExecution("""
            import sys
            print(sys._getframe())
            """)

    ############################################################
    # _mercurial
    @expectedFailure
    def test__mercurial(self):
        self.assertCodeExecution("""
            import sys
            print(sys._mercurial)
            """)

    ############################################################
    # _xoptions
    @expectedFailure
    def test__xoptions(self):
        self.assertCodeExecution("""
            import sys
            print(sys._xoptions)
            """)

    ############################################################
    # abiflags
    @expectedFailure
    def test_abiflags(self):
        self.assertCodeExecution("""
            import sys
            print(sys.abiflags)
            """)

    ############################################################
    # api_version
    @expectedFailure
    def test_api_version(self):
        self.assertCodeExecution("""
            import sys
            print(sys.api_version)
            """)

    ############################################################
    # argv
    def test_argv(self):
        self.assertCodeExecution("""
            import sys
            print('ARGS =', sys.argv)
            """)

        self.assertCodeExecution("""
            import sys
            print('ARGS =', sys.argv)
            """, args=['1', 'asdf', '3'])

    ############################################################
    # base_exec_prefix
    @expectedFailure
    def test_base_exec_prefix(self):
        self.assertCodeExecution("""
            import sys
            print(sys.base_exec_prefix)
            """)

    ############################################################
    # base_prefix
    @expectedFailure
    def test_base_prefix(self):
        self.assertCodeExecution("""
            import sys
            print(sys.base_prefix)
            """)

    ############################################################
    # builtin_module_names
    @expectedFailure
    def test_builtin_module_names(self):
        self.assertCodeExecution("""
            import sys
            print(sys.builtin_module_names)
            """)

    ############################################################
    # byteorder
    @expectedFailure
    def test_byteorder(self):
        self.assertCodeExecution("""
            import sys
            print(sys.byteorder)
            """)

    ############################################################
    # call_tracing
    @expectedFailure
    def test_call_tracing(self):
        self.assertCodeExecution("""
            import sys
            print(sys.call_tracing())
            """)

    ############################################################
    # callstats
    @expectedFailure
    def test_callstats(self):
        self.assertCodeExecution("""
            import sys
            print(sys.callstats())
            """)

    ############################################################
    # copyright
    @expectedFailure
    def test_copyright(self):
        self.assertCodeExecution("""
            import sys
            print(sys.copyright)
            """)

    ############################################################
    # displayhook
    @expectedFailure
    def test_displayhook(self):
        self.assertCodeExecution("""
            import sys
            print(sys.displayhook)
            """)

    ############################################################
    # dont_write_bytecode
    @expectedFailure
    def test_dont_write_bytecode(self):
        self.assertCodeExecution("""
            import sys
            print(sys.dont_write_bytecode)
            """)

    ############################################################
    # exc_info
    @expectedFailure
    def test_exc_info(self):
        self.assertCodeExecution("""
            import sys
            prexc_info()int(sys.exc_info())
            """)

    ############################################################
    # excepthook
    @expectedFailure
    def test_excepthook(self):
        self.assertCodeExecution("""
            import sys
            print(sys.excepthook)
            """)

    ############################################################
    # exec_prefix
    @expectedFailure
    def test_exec_prefix(self):
        self.assertCodeExecution("""
            import sys
            print(sys.exec_prefix)
            """)

    ############################################################
    # executable
    @expectedFailure
    def test_executable(self):
        self.assertCodeExecution("""
            import sys
            print(sys.executable)
            """)

    ############################################################
    # exit
    def test_exit_from_static(self):
        # Exit from static block
        self.assertCodeExecution("""
            import sys
            print(sys.exit())
            """, exits_early=True)

    def test_exit(self):
        # From inside main
        self.assertCodeExecution("""
            import sys
            if __name__ == '__main__':
                print(sys.exit())
            """, run_in_function=False, exits_early=True)

        # From a method called from inside main
        self.assertCodeExecution("""
            import sys

            def foo():
                print(sys.exit())

            if __name__ == '__main__':
                foo()
            """, run_in_function=False, exits_early=True)

    ############################################################
    # flags
    @expectedFailure
    def test_flags(self):
        self.assertCodeExecution("""
            import sys
            print(sys.flags())
            """)

    ############################################################
    # float_info
    @expectedFailure
    def test_float_info(self):
        self.assertCodeExecution("""
            import sys
            print(sys.float_info())
            """)

    ############################################################
    # float_repr_style
    @expectedFailure
    def test_float_repr_style(self):
        self.assertCodeExecution("""
            import sys
            print(sys.float_repr_style)
            """)

    ############################################################
    # getallocatedblocks
    @expectedFailure
    def test_getallocatedblocks(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getallocatedblocks())
            """)

    ############################################################
    # getcheckinterval
    @expectedFailure
    def test_getcheckinterval(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getcheckinterval())
            """)

    ############################################################
    # getdefaultencoding
    @expectedFailure
    def test_getdefaultencoding(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getdefaultencoding())
            """)

    ############################################################
    # getdlopenflags
    @expectedFailure
    def test_getdlopenflags(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getdlopenflags())
            """)

    ############################################################
    # getfilesystemencoding
    @expectedFailure
    def test_getfilesystemencoding(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getfilesystemencoding())
            """)

    ############################################################
    # getprofile
    @expectedFailure
    def test_getprofile(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getprofile())
            """)

    ############################################################
    # getrecursionlimit
    @expectedFailure
    def test_getrecursionlimit(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getrecursionlimit())
            """)

    ############################################################
    # getrefcount
    @expectedFailure
    def test_getrefcount(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getrefcount())
            """)

    ############################################################
    # getsizeof
    @expectedFailure
    def test_getsizeof(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getsizeof())
            """)

    ############################################################
    # Float getswitchinterval
    @expectedFailure
    def test_getswi(self):
        self.assertCodeExecution("""
            import sys
            print(sys.getswi())
            """)

    ############################################################
    # gettrace
    @expectedFailure
    def test_gettrace(self):
        self.assertCodeExecution("""
            import sys
            print(sys.gettrace())
            """)

    ############################################################
    # hash_info
    @expectedFailure
    def test_hash_info(self):
        self.assertCodeExecution("""
            import sys
            print(sys.hash_info)
            """)

    ############################################################
    # hexversion
    def test_hexversion(self):
        self.assertCodeExecution("""
            import sys
            print(sys.hexversion)
            """)

    ############################################################
    # implementation
    @expectedFailure
    def test_implementati(self):
        self.assertCodeExecution("""
            import sys
            print(sys.implementati())
            """)

    ############################################################
    # int_info
    @expectedFailure
    def test_int_info(self):
        self.assertCodeExecution("""
            import sys
            print(sys.int_info())
            """)

    ############################################################
    # intern
    @expectedFailure
    def test_intern(self):
        self.assertCodeExecution("""
            import sys
            print(sys.intern())
            """)

    ############################################################
    # last_traceback
    @expectedFailure
    def test_last_traceba(self):
        self.assertCodeExecution("""
            import sys
            print(sys.last_traceba())
            """)

    ############################################################
    # last_type
    @expectedFailure
    def test_last_type(self):
        self.assertCodeExecution("""
            import sys
            print(sys.last_type())
            """)

    ############################################################
    # last_value
    @expectedFailure
    def test_last_value(self):
        self.assertCodeExecution("""
            import sys
            print(sys.last_value())
            """)

    ############################################################
    # maxsize
    def test_maxsize(self):
        self.assertCodeExecution("""
            import sys
            print(sys.maxsize)
            """)

    ############################################################
    # maxunicode
    @expectedFailure
    def test_maxunicode(self):
        self.assertCodeExecution("""
            import sys
            print(sys.maxunicode)
            """)

    ############################################################
    # meta_path
    @expectedFailure
    def test_meta_path(self):
        self.assertCodeExecution("""
            import sys
            print(sys.meta_path)
            """)

    ############################################################
    # modules
    @expectedFailure
    def test_modules(self):
        self.assertCodeExecution("""
            import sys
            print(sys.modules)
            """)

    ############################################################
    # path
    @expectedFailure
    def test_path(self):
        self.assertCodeExecution("""
            import sys
            print(sys.path)
            """)

    ############################################################
    # path_hooks
    @expectedFailure
    def test_path_hooks(self):
        self.assertCodeExecution("""
            import sys
            print(sys.path_hooks)
            """)

    ############################################################
    # path_importer_cache
    @expectedFailure
    def test_path_importer_cache(self):
        self.assertCodeExecution("""
            import sys
            print(sys.path_importer_cache)
            """)

    ############################################################
    # platform
    def test_platform(self):
        self.assertCodeExecution("""
            import sys
            print(sys.platform)
            """)

    ############################################################
    # prefix
    @expectedFailure
    def test_prefix(self):
        self.assertCodeExecution("""
            import sys
            print(sys.prefix)
            """)

    ############################################################
    # ps1
    @expectedFailure
    def test_ps1(self):
        self.assertCodeExecution("""
            import sys
            print(sys.ps1)
            """)

    ############################################################
    # ps2
    @expectedFailure
    def test_ps2(self):
        self.assertCodeExecution("""
            import sys
            print(sys.ps2)
            """)

    ############################################################
    # ps3
    @expectedFailure
    def test_ps3(self):
        self.assertCodeExecution("""
            import sys
            print(sys.ps3)
            """)

    ############################################################
    # real_prefix
    @expectedFailure
    def test_real_prefix(self):
        self.assertCodeExecution("""
            import sys
            print(sys.real_prefix)
            """)

    ############################################################
    # setcheckinterval
    @expectedFailure
    def test_setcheckinterval(self):
        self.assertCodeExecution("""
            import sys
            print(sys.setcheckinterval())
            """)

    ############################################################
    # setdlopenflags
    @expectedFailure
    def test_setdlopenflags(self):
        self.assertCodeExecution("""
            import sys
            print(sys.setdlopenflags())
            """)

    ############################################################
    # setprofile
    @expectedFailure
    def test_setprofile(self):
        self.assertCodeExecution("""
            import sys
            print(sys.setprofile())
            """)

    ############################################################
    # setrecursionlimit
    @expectedFailure
    def test_setrecursionlimit(self):
        self.assertCodeExecution("""
            import sys
            print(sys.setrecursionlimit())
            """)

    ############################################################
    # setswitchinterval
    @expectedFailure
    def test_setswitchinterval(self):
        self.assertCodeExecution("""
            import sys
            print(sys.setswitchinterval())
            """)

    ############################################################
    # settrace
    @expectedFailure
    def test_settrace(self):
        self.assertCodeExecution("""
            import sys
            print(sys.settrace())
            """)

    ############################################################
    # stderr
    @expectedFailure
    def test_stderr(self):
        self.assertCodeExecution("""
            import sys
            print(sys.stderr())
            """)

    ############################################################
    # stdin
    @expectedFailure
    def test_stdin(self):
        self.assertCodeExecution("""
            import sys
            print(sys.stdin())
            """)

    ############################################################
    # stdout
    @expectedFailure
    def test_stdout(self):
        self.assertCodeExecution("""
            import sys
            print(sys.stdout())
            """)

    ############################################################
    # thread_info
    @expectedFailure
    def test_thread_info(self):
        self.assertCodeExecution("""
            import sys
            print(sys.thread_info())
            """)

    ############################################################
    # version
    @expectedFailure
    def test_version(self):
        self.assertCodeExecution("""
            import sys
            lines = sys.version.split('\n')
            print("There are %d lines" % len(lines))
            print(lines[0].split(' ')[0])
            """)

    ############################################################
    # version_info
    def test_version_info(self):
        self.assertCodeExecution("""
            import sys
            print(sys.version_info)
            """)

    ############################################################
    # warnoptions
    @expectedFailure
    def test_warnoptions(self):
        self.assertCodeExecution("""
            import sys
            print(sys.warnoptions)
            """)



from .. utils import TranspileTestCase, BuiltinTwoargFunctionTestCase


class HasattrTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            class MyClass(object):
                class_value = 42

                def __init__(self, val):
                    self.value = val

                def stuff(self, delta):
                    print("DELTA: ", delta)
                    return self.value + delta

            print("On class: ")
            print('MyClass.foo', hasattr(MyClass, 'foo'))
            print('MyClass.value', hasattr(MyClass, 'value'))
            print('MyClass.class_value', hasattr(MyClass, 'class_value'))
            print('MyClass.stuff', hasattr(MyClass, 'stuff'))

            obj = MyClass(37)

            print("On instance:")
            print('obj.foo', hasattr(obj, 'foo'))
            print('obj.value', hasattr(obj, 'value'))
            print('obj.class_value', hasattr(obj, 'class_value'))
            print('obj.stuff', hasattr(obj, 'stuff'))
            """, run_in_function=False)


class BuiltinHasattrFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["hasattr"]

    not_implemented = [
    ]



import unittest

from ..utils import TranspileTestCase


class OsModuleTests(TranspileTestCase):
    def test_getenv(self):
        self.assertCodeExecution("""
            import os
            print(os.getenv('USER'))
            print(os.getenv('NOT_HERE', default='something'))
            print(os.getenv('NOT_HERE', 'something'))
            print(os.getenv('NOT_HERE'))
            """)
        self.assertCodeExecution("""
            import os
            try:
                print(os.getenv(None))
            except Exception as e:
                print(type(e), e)
            try:
                print(os.getenv(1))
            except Exception as e:
                print(type(e), e)
            try:
                print(os.getenv())
            except Exception as e:
                print(type(e), e)
            """)

    @unittest.expectedFailure
    def test_getcwd(self):
        self.assertCodeExecution("""
            import os
            print(os.getcwd())
        """)

    def test_cpu_count(self):
        self.assertCodeExecution("""
            import os
            print(os.cpu_count())
        """)



from unittest import expectedFailure

from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class TypeTests(TranspileTestCase):

    @expectedFailure
    def test_dynamic_class_definitions_leak_state(self):
        self.assertCodeExecution("""
            my_class_obj = type("MyClass", (object,), {})
            print("my_class_obj: ", my_class_obj)

            obj = object
            print("obj: ", object)

            dynamic_obj = type("object", (object,), {})
            print("dynamic_obj: ", dynamic_obj)

            # Check state leak
            print("my_class_obj: ", my_class_obj)
            print("obj: ", obj)
        """)


class BuiltinTypeFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["type"]

    not_implemented = [
    ]



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class NoneTypeTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = None
            try:
                x.thing = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = None
            try:
                y = x.thing
            except AttributeError as err:
                print(err)
            """)


class UnaryNoneTypeOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'None'


class BinaryNoneTypeOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'None'


class InplaceNoneTypeOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'None'



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class ReprTests(TranspileTestCase):
    pass


class BuiltinReprFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["repr"]

    not_implemented = [
        'test_class',
        'test_complex',
    ]



from unittest import expectedFailure

from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class TupleTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = (1, 2, 3)
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = (1, 2, 3)
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_creation(self):
        self.assertCodeExecution("""
            a = 1
            b = 2
            c = 3
            d = 4
            e = 5
            x = (a, b, c, d, e)
            print(x)
            """)

    def test_const_creation(self):
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            print(x)
            """)

    def test_const_creation_multitype(self):
        self.assertCodeExecution("""
            x = (1, 2.5, "3", True, 5)
            print(x)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            try:
                x[0] = "a"
            except TypeError as err:
                print(err)
        """)

    def test_getitem(self):
        # Simple positive index
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            print(x[2])
            """)

        # Simple negative index
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            print(x[-2])
            """)

        # Positive index out of range
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            try:
                print(x[10])
            except IndexError as err:
                print(err)
            """)

        # Negative index out of range
        self.assertCodeExecution("""
            x = (1, 2, 3, 4, 5)
            try:
                print(x[-10])
            except IndexError as err:
                print(err)
            """)

    def test_lt(self):
        self.assertCodeExecution("""
            pairs = [
                (1, 2, 3), (3, 2, 1),
                (1, 2, 1), (1, 2, 4),
                (1, 2, 3, 4), (1, 2, 3, 0)
            ]
            for p1 in pairs:
                for p2 in pairs:
                    print(p1, "<", p2, ":", p1 < p2)
            """, run_in_function=False)

    def test_immutable(self):
        self.assertCodeExecution("""
            x = (1, 2)
            try:
                x[0] = 3
            except TypeError as err:
                print(err)
            try:
                x[-1] = 3
            except TypeError as err:
                print(err)
            try:
                x[2] = 3
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_lt_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __gt__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __lt__()
            print(x.__lt__(x))

            # ensure rich comparison logic is used
            print((x,) < (x,)) # False, x is x and same size
            print((x,) < (y,)) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print((x, y) < (y,)) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print((x, y) < (x,)) # False, larger size
            print((x,) < (x, y)) # True, smaller size
            """)

    def test_le_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __ge__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __le__()
            print(x.__le__(x))

            # ensure rich comparison logic is used
            print((x,) <= (x,)) # False, x is x and same size
            print((x,) <= (y,)) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print((x, y) <= (y,)) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print((x, y) <= (x,)) # False, larger size
            print((x,) <= (x, y)) # True, smaller size
            """)

    def test_gt_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __lt__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __gt__()
            print(x.__gt__(x))

            # ensure rich comparison logic is used
            print((x,) > (x,)) # False, x is x and same size
            print((x,) > (y,)) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print((x, y) > (y,)) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print((x, y) > (x,)) # False, larger size
            print((x,) > (x, y)) # True, smaller size
            """)

    def test_ge_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __le__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __ge__()
            print(x.__ge__(x))

            # ensure rich comparison logic is used
            print((x,) >= (x,)) # False, x is x and same size
            print((x,) >= (y,)) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print((x, y) >= (y,)) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print((x, y) >= (x,)) # False, larger size
            print((x,) >= (x, y)) # True, smaller size
            """)

    def test_eq_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __eq__(self, other):
                    return True

            class B:
                def __eq__(self, other):
                    return False

            x = A()
            y = B()

            print((x,) == (x,)) # True, identity implies equality
            print((x, x) == (x,)) # False, size not equal

            print((x,) == (y,)) # True, x is not y, x.__eq__(y)
            print((y,) == (x,)) # False, y is not x, y.__eq__(x)
            """)

    def test_iteration(self):
        self.assertCodeExecution("""
            a = (1, 2, 3, 4, 5)
            b = ()
            for x in a:
                print(x)
            for x in b:
                print(x)
            """)

    def test_index(self):
        # success
        self.assertCodeExecution("""
            a = (1, 2, 3, 4, 3)
            print(a.index(2))
            print(a.index(2,1))
            print(a.index(2,1,5))
            print(a.index(2,0,1000))
            print(a.index(2,-300))
            """)

        # raise exceptions
        self.assertCodeExecution("""
            a = (1, 2, 3, 4, 3)
            try:
                print(a.index('2'))
            except ValueError as err:
                print(err)
            try:
                print(a.index(2,0,-50))
            except ValueError as err:
                print(err)
            try:
                print(a.index(7,0,5))
            except ValueError as err:
                print(err)
            try:
                print(a.index())
            except TypeError as err:
                print(err)
            """)

    def test_no_arguments(self):
        self.assertCodeExecution("""
            x = tuple()
            print(x)
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(tuple(0, 1))
            except TypeError as err:
                print(err)
        """)

    def test_wrong_argument(self):
        self.assertCodeExecution("""
            try:
                print(tuple(0))
            except TypeError as err:
                print(err)
        """)


class UnaryTupleOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'tuple'


class BinaryTupleOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'tuple'


class InplaceTupleOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'tuple'



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class BoolTests(TranspileTestCase):
    def test_bool_omitted(self):
        self.assertCodeExecution("""
            print(bool())
            """)

    def test_bool_like(self):
        self.assertCodeExecution("""
            class BoolLike:
                def __init__(self, val):
                    self.val = val

                def __bool__(self):
                    return self.val == 1

            print(bool(BoolLike(0)))
            print(bool(BoolLike(1)))
            """)

    def test_len_only(self):
        self.assertCodeExecution("""
            class LenButNoBool:
                def __init__(self, val):
                    self.val = val

                def __len__(self):
                    return self.val

            print(bool(LenButNoBool(0)))
            print(bool(LenButNoBool(1)))
            """)

    def test_no_bool_no_len(self):
        self.assertCodeExecution("""
            class NoLenNoBool:
                def __init__(self, val):
                    self.val = val

            print(bool(NoLenNoBool(0)))
            print(bool(NoLenNoBool(1)))
            print(bool(NoLenNoBool(42)))
            print(bool(NoLenNoBool(-2)))
            """)


class BuiltinBoolFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["bool"]

    not_implemented = [
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class ObjectTests(TranspileTestCase):
    pass


class BuiltinObjectFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["object"]

    not_implemented = [
    ]



from unittest import expectedFailure

from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class SetTests(TranspileTestCase):
    @expectedFailure
    def test_complex_element(self):
        self.assertCodeExecution("""
            x = {1j, 2j}

            for i in x:
                print(i)
            """)

    @expectedFailure
    def test_equal_value_different_datatypes(self):
        self.assertCodeExecution("""
            x = {1, 1.0, True}

            print("size of set is:", len(x))

            for i in x:
                print(i)
            """)

    def test_setattr(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_creation(self):
        # Empty dict
        self.assertCodeExecution("""
            x = set()
            print(x)
            """)

        # Set constant
        self.assertCodeExecution("""
            x = {'a'}
            print(x)
            """)

        self.assertCodeExecution("""
            x = set(['a'])
            print(x)
            """)

    def test_getitem(self):
        # Simple existent key
        self.assertCodeExecution("""
            x = {'a', 'b'}
            print('a' in x)
            """)

        # Simple non-existent key
        self.assertCodeExecution("""
            x = {'a', 'b'}
            print('c' in x)
            """)

    def test_pop(self):
        # Test empty set popping.
        self.assertCodeExecution("""
            x = set()
            try:
                x.pop()
            except KeyError as err:
                print(err)
        """)

        # Test populated set popping.
        self.assertCodeExecution("""
            x = {'a'}
            print(len(x) == 1)
            x.pop()
            print(len(x) == set())
        """)

        # Test popping returns from set.
        self.assertCodeExecution("""
            x = {'a', 'b', 'c', 'd', 'e'}
            y = x.pop()
            print(y in x)
        """)

    def test_copy(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = x.copy()
            print(x == y)
            print(x is not y)
            """)

        # ensures that it did a shallow copy
        self.assertCodeExecution("""
            x = {(1, 2, 3)}
            y = x.copy()
            print(x.pop() is y.pop())
            """)

    def test_difference(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            z = x.difference(y)
            w = x.difference([3,4,5])
            print(x)
            print(y)
            print(z)
            print(w)
            """)

        # not iterable test
        self.assertCodeExecution("""
            x = set([1, 2, 3])
            try:
                print(x.difference(1))
            except TypeError as err:
                print(err)
            """)

    def test_discard(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            x.discard(1)
            x.discard(4)
            print(x)
            """)

    def test_intersection(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            z = x.intersection(y)
            w = x.intersection([3,4,5])
            print(x)
            print(y)
            print(z)
            print(w)
            """)

        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {4, 5}
            z = x.intersection(y)
            print(x)
            print(y)
            print(z)
            """)

        # not iterable test
        self.assertCodeExecution("""
            x = set([1, 2, 3])
            try:
                print(x.intersection(1))
            except TypeError as err:
                print(err)
            """)

    def test_remove(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            x.remove(1)
            try:
                x.remove(4)
            except KeyError as err:
                print(err)
            print(x)
            """)

    def test_union(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            z = x.union(y)
            print(x)
            print(y)
            print(z)
            """)

    def test_difference_update(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            x.difference_update(y)
            print(x)
            """)

    def test_intersection_update(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            x.intersection_update(y)
            print(x)
            """)

    def test_update(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = {3, 4, 5}
            print(x.update(y))
            print(x)
            """)

    def test_iteration(self):
        self.assertCodeExecution("""
            x = {1, 2, 3}
            y = set()
            #We are not printing each element because python and voc store the items in different ways.
            for s in x:
                print(x.__contains__(s))
            for s in y:
                print(y.__conatins__(s))
            """)

    def test_isdisjoint(self):
        # successful
        self.assertCodeExecution("""
            x = set("hello World !")
            y = set("hello")
            z = set()
            t = set()
            print(x.isdisjoint(y))
            print(x.isdisjoint("hell"))
            print(x.isdisjoint(z))
            print(t.isdisjoint(z))
            print(t.isdisjoint(y))
            """)

        # unsuccessful
        self.assertCodeExecution("""
            x = 9.3
            y = 5
            l = set("hello World !")
            m = set("hello")
            try:
                print(l.isdisjoint(y))
            except TypeError as err:
                print(err)
            try:
                print(l.isdisjoint(x))
            except TypeError as err:
                print(err)
            try:
                print(l.isdisjoint(x))
            except TypeError as err:
                print(err)
            """)

    def test_issubset(self):
        self.assertCodeExecution("""
            a = set('abc')
            b = set('abcde')
            c = set()
            print(a.issubset(b))
            print(a.issubset('ab'))
            print(a.issubset(a))
            print(a.issubset(c))
            """)

        # not iterable test
        self.assertCodeExecution("""
            a = {1, 2, 3}
            try:
                print(a.issubset(1))
            except TypeError as err:
                print(err)
            """)

    def test_issuperset(self):
        self.assertCodeExecution("""
            a = set('abcd')
            b = set('ab')
            c = set()
            print(a.issuperset(b))
            print(a.issuperset('ab1'))
            print(a.issuperset(a))
            print(a.issuperset(c))
            """)

        # not iterable test
        self.assertCodeExecution("""
            a = {1, 2, 3}
            try:
                print(a.issuperset(1))
            except TypeError as err:
                print(err)
            """)

    def test_symmetric_difference(self):
        # successful
        self.assertCodeExecution("""
            x = set("hello World !")
            y = set("hello")
            z = set()
            t = set()
            print(sorted(x.symmetric_difference(y)))
            print(sorted(x.symmetric_difference(z)))
            print(sorted(t.symmetric_difference(z)))
            print(sorted(t.symmetric_difference(y)))
            """)

        # unsuccessful
        self.assertCodeExecution("""
            x = 9.3
            y = 5
            l = set("hello World !")
            m = set("hello")
            try:
                print(l.symmetric_difference(y))
            except TypeError as err:
                print(err)
            try:
                print(l.symmetric_difference(x))
            except TypeError as err:
                print(err)
            try:
                print(l.symmetric_difference())
            except TypeError as err:
                print(err)
            """)

    def test_symmetric_difference_update(self):
        # successful
        self.assertCodeExecution("""
            x = set("hello World !")
            y = set("hello")
            z = set()
            t = set()
            x.symmetric_difference_update(y)
            print(sorted(x))
            #x.test_symmetric_difference_update(z)
            print(sorted(x))
            t.symmetric_difference_update(z)
            print(sorted(t))
            t.symmetric_difference_update(y)
            print(sorted(t))
            """)

        # unsuccessful
        self.assertCodeExecution("""
            x = 9.3
            y = 5
            l = set("hello World !")
            m = set("hello")
            try:
                print(l.symmetric_difference_update(y))
            except TypeError as err:
                print(err)
            try:
                print(l.symmetric_difference_update(x))
            except TypeError as err:
                print(err)
            try:
                print(l.symmetric_difference_update())
            except TypeError as err:
                print(err)
            """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(set(1, 2))
            except TypeError as err:
                print(err)
        """)


class UnarySetOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'set'


class BinarySetOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'set'


class InplaceSetOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'set'



import os

from .utils import adjust, TranspileTestCase


TESTDIR = os.path.dirname(__file__)


class BenchmarkTests(TranspileTestCase):
    def test_binary_trees(self):
        with open(os.path.join(TESTDIR + "/benchmarks/", "binary-trees.py")) as javafile:
            out = self.runAsJava(adjust(javafile.read()), args=["5", "5"], timed=True)

            self.assertIn("stretch tree of depth", out)

    def test_call_simple(self):
        with open(os.path.join(TESTDIR + "/benchmarks/", "call_simple.py")) as javafile:
            out = self.runAsJava(adjust(javafile.read()), args=["2"])

            self.assertIn("Time elapsed: ", out)

    def test_unpack_sequence(self):
        with open(os.path.join(TESTDIR + "/benchmarks/", "unpack_sequence.py")) as javafile:
            out = self.runAsJava(adjust(javafile.read()), args=["10"])

            self.assertIn("Time elapsed: ", out)

    def test_pystone(self):
        with open(os.path.join(TESTDIR + "/benchmarks/", "pystone.py")) as javafile:
            out = self.runAsJava(adjust(javafile.read()), args=["10"])

            self.assertIn("Pystone(1.2) time for 10 passes = ", out)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase, SAMPLE_SUBSTITUTIONS


class TupleTests(TranspileTestCase):
    pass


class BuiltinTupleFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["tuple"]

    not_implemented = [
        'test_tuple',
    ]

    substitutions = {
        "('one', 'two', 'six')": [
            "('two', 'one', 'six')", "('six', 'one', 'two')", "('one', 'six', 'two')", "('two', 'six', 'one')", "('six', 'two', 'one')"
        ],
        "('on', 'to', 'an')": [
            "('to', 'on', 'an')", "('an', 'on', 'to')", "('on', 'an', 'to')", "('to', 'an', 'on')", "('an', 'to', 'on')"
        ],
        "(1, 2.3456, 7)": [
            "(2.3456, 1, 7)", "(7, 1, 2.3456)", "(1, 7, 2.3456)", "(2.3456, 7, 1)", "(7, 2.3456, 1)"
        ],
        "('a', 'c', 'd')": [
            "('c', 'a', 'd')", "('d', 'a', 'c')", "('a', 'd', 'c')", "('c', 'd', 'a')", "('d', 'c', 'a')"
        ]
    }

    substitutions.update(SAMPLE_SUBSTITUTIONS)



from .attributes import Attribute
from .constants import Utf8


# From: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html


##########################################################################
# 4.5. Fields
##########################################################################

# Each field is described by a field_info structure. No two fields in one class
# file may have the same name and descriptor (§4.3.2).

# The structure has the following format:

class Field:
    # u2 access_flags;
    # u2 name_index;
    # u2 descriptor_index;
    # u2 attributes_count;
    # attribute_info attributes[attributes_count];

    # Table 4.4. Field access and property flags

    # Flag Name   Value   Interpretation
    ACC_PUBLIC = 0x0001     # Declared public; may be accessed from outside its package.
    ACC_PRIVATE = 0x0002    # Declared private; usable only within the defining class.
    ACC_PROTECTED = 0x0004  # Declared protected; may be accessed within subclasses.
    ACC_STATIC = 0x0008     # Declared static.
    ACC_FINAL = 0x0010      # Declared final; never directly assigned to after object construction (JLS §17.5).
    ACC_VOLATILE = 0x0040   # Declared volatile; cannot be cached.
    ACC_TRANSIENT = 0x0080  # Declared transient; not written or read by a persistent object manager.
    ACC_SYNTHETIC = 0x1000  # Declared synthetic; not present in the source code.
    ACC_ENUM = 0x4000       # Declared as an element of an enum.

    def __init__(
                self, name, descriptor,
                public=True, private=False, protected=False,
                static=False, final=False, volatile=False, transient=False,
                synthetic=False, enum=False,
                attributes=None
            ):

        # A field may be marked with the ACC_SYNTHETIC flag to indicate that it
        # was generated by a compiler and does not appear in source code.
        #
        # The ACC_ENUM flag indicates that this field is used to hold an element
        # of an enumerated type.
        #
        # Fields of classes may set any of the flags in Table 4.4. However, a
        # specific field of a class may have at most one of its ACC_PRIVATE,
        # ACC_PROTECTED, and ACC_PUBLIC flags set (JLS §8.3.1) and must not have
        # both its ACC_FINAL and ACC_VOLATILE flags set (JLS §8.3.1.4).
        #
        # All fields of interfaces must have their ACC_PUBLIC, ACC_STATIC, and
        # ACC_FINAL flags set; they may have their ACC_SYNTHETIC flag set and
        # must not have any of the other flags in Table 4.4 set (JLS §9.3).

        self.private = private
        if self.private:
            self.protected = False
            self.public = False
        else:
            self.protected = protected
            if self.protected:
                self.public = False
            else:
                self.public = True

        self.static = static
        self.final = final
        self.volatile = volatile
        self.transient = transient
        self.synthetic = synthetic
        self.enum = enum

        # The value of the name_index item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure which must represent a valid
        # unqualified name (§4.2.2) denoting a field.
        self.name = Utf8(name)

        # The value of the descriptor_index item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure that must represent a valid
        # field descriptor (§4.3.2).
        self.descriptor = Utf8(descriptor)

        # Each value of the attributes table must be an attribute structure
        # (§4.7). A field can have any number of attributes associated with it.
        #
        # The attributes defined by this specification as appearing in the
        # attributes table of a field_info structure are ConstantValue (§4.7.2),
        # Synthetic (§4.7.8), Signature (§4.7.9), Deprecated (§4.7.15),
        # RuntimeVisibleAnnotations (§4.7.16) and RuntimeInvisibleAnnotations
        # (§4.7.17).
        #
        # A Java Virtual Machine implementation must recognize and correctly
        # read ConstantValue (§4.7.2) attributes found in the attributes table
        # of a field_info structure. If a Java Virtual Machine implementation
        # recognizes class files whose version number is 49.0 or above, it must
        # recognize and correctly read Signature (§4.7.9),
        # RuntimeVisibleAnnotations (§4.7.16) and RuntimeInvisibleAnnotations
        # (§4.7.17) attributes found in the attributes table of a field_info
        # structure of a class file whose version number is 49.0 or above.
        #
        # A Java Virtual Machine implementation is required to silently ignore
        # any or all attributes that it does not recognize in the attributes
        # table of a field_info structure. Attributes not defined in this
        # specification are not allowed to affect the semantics of the class
        # file, but only to provide additional descriptive information (§4.7.1).
        self.attributes = attributes if attributes else []

    def __repr__(self):
        return '<Field access:0x%04x name:%s, descriptor:%s>' % (self.access_flags, self.name, self.descriptor)

    @staticmethod
    def read(reader, dump=None):
        access_flags = reader.read_u2()

        name = reader.constant_pool[reader.read_u2()].bytes.decode('mutf-8')
        descriptor = reader.constant_pool[reader.read_u2()].bytes.decode('mutf-8')
        attributes_count = reader.read_u2()

        if dump:
            reader.debug("    " * dump, 'Field %s (%s)' % (name, descriptor))

            access_description = ', '.join(f for f in [
                    flag if access_flags & mask else None
                    for flag, mask in [
                        ('public', Field.ACC_PUBLIC),
                        ('private', Field.ACC_PRIVATE),
                        ('protected', Field.ACC_PROTECTED),
                        ('static', Field.ACC_STATIC),
                        ('final', Field.ACC_FINAL),
                        ('volatile', Field.ACC_VOLATILE),
                        ('transient', Field.ACC_TRANSIENT),
                        ('synthetic', Field.ACC_SYNTHETIC),
                        ('enum', Field.ACC_ENUM),
                    ]
                ] if f)
            reader.debug("    " * dump, '    Flags: 0x%04x%s' % (access_flags, ' (%s)') % (
                access_description if access_description else ''
            ))

            reader.debug("    " * dump, '    Attributes: (%s)' % attributes_count)

        attributes = []
        for i in range(0, attributes_count):
            attributes.append(
                Attribute.read(reader, dump=dump + 2 if dump is not None else dump)
            )

        return Field(
            name=name,
            descriptor=descriptor,
            public=bool(access_flags & Field.ACC_PUBLIC),
            private=bool(access_flags & Field.ACC_PRIVATE),
            protected=bool(access_flags & Field.ACC_PROTECTED),
            static=bool(access_flags & Field.ACC_STATIC),
            final=bool(access_flags & Field.ACC_FINAL),
            volatile=bool(access_flags & Field.ACC_VOLATILE),
            transient=bool(access_flags & Field.ACC_TRANSIENT),
            synthetic=bool(access_flags & Field.ACC_SYNTHETIC),
            enum=bool(access_flags & Field.ACC_ENUM),
            attributes=attributes,
        )

    def write(self, writer):
        writer.write_u2(self.access_flags)
        writer.write_u2(writer.constant_pool.index(self.name))
        writer.write_u2(writer.constant_pool.index(self.descriptor))
        writer.write_u2(self.attributes_count)

        for attribute in self.attributes:
            attribute.write(writer)

    def resolve(self, constant_pool):
        constant_pool.add(self.name)
        constant_pool.add(self.descriptor)

        for attribute in self.attributes:
            attribute.resolve(constant_pool)

    @property
    def attributes_count(self):
        """The value of the attributes_count item indicates the number of
        additional attributes (§4.7) of this field.
        """
        return len(self.attributes)

    @property
    def access_flags(self):
        """The value of the access_flags item is a mask of flags used to denote
        access permission to and properties of this field. The interpretation
        of each flag, when set, is as shown in Table 4.4.

        All bits of the access_flags item not assigned in Table 4.4 are
        reserved for future use. They should be set to zero in generated class
        files and should be ignored by Java Virtual Machine implementations.
        """
        return (
            (self.ACC_PUBLIC if self.public else 0) |
            (self.ACC_PRIVATE if self.private else 0) |
            (self.ACC_PROTECTED if self.protected else 0) |
            (self.ACC_STATIC if self.static else 0) |
            (self.ACC_FINAL if self.final else 0) |
            (self.ACC_VOLATILE if self.volatile else 0) |
            (self.ACC_TRANSIENT if self.transient else 0) |
            (self.ACC_SYNTHETIC if self.synthetic else 0) |
            (self.ACC_ENUM if self.enum else 0)
        )



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class AllTests(TranspileTestCase):
    pass


class BuiltinAllFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["all"]



from .. utils import TranspileTestCase


class IsinstanceTests(TranspileTestCase):
    def test_no_args(self):
        self.assertCodeExecution("""
            try:
                print(isinstance())
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_one_arg(self):
        self.assertCodeExecution("""
            try:
                print(isinstance(123))
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_non_tuple_second_arg(self):
        self.assertCodeExecution("""
            try:
                print(isinstance(123, [int, float]))
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_single_class(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2:
                pass

            obj = MyClass1()

            print(isinstance(obj, MyClass1))
            print(isinstance(obj, MyClass2))
            """, run_in_function=False)

    def test_multiple_classes(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2:
                pass

            obj = MyClass1()

            print(isinstance(obj, (MyClass1, MyClass2)))
            """, run_in_function=False)

    def test_parent(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2(MyClass1):
                pass

            obj = MyClass2()

            print(isinstance(obj, MyClass1))
            print(isinstance(obj, MyClass2))
            """, run_in_function=False)

    def test_grandparent(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2(MyClass1):
                pass

            class MyClass3(MyClass2):
                pass

            obj = MyClass3()

            print(isinstance(obj, MyClass1))
            print(isinstance(obj, MyClass2))
            print(isinstance(obj, MyClass3))
            """, run_in_function=False)

    def test_types(self):
        self.assertCodeExecution("""
            print(isinstance(123, int))
            print(isinstance(123, str))
            print(isinstance(123, float))

            print(isinstance("asdf", int))
            print(isinstance("asdf", str))
            print(isinstance("asdf", float))

            print(isinstance(123.45, int))
            print(isinstance(123.45, str))
            print(isinstance(123.45, float))

            print(isinstance(123, (int, float)))
            print(isinstance("asdf", (int, float)))
            print(isinstance(123.45, (int, float)))
            """, run_in_function=False)



import os

from ..java import (
    Annotation,
    Class as JavaClass,
    Classref as JavaClassref,
    Code as JavaCode,
    ConstantElementValue,
    Field as JavaField,
    Method as JavaMethod, RuntimeVisibleAnnotations,
    SourceFile,
    opcodes as JavaOpcodes,
)
from .blocks import Block, IgnoreBlock
from .methods import (
    InitMethod, ClosureInitMethod,
    GeneratorMethod, Method, CO_GENERATOR,
)
from .types import java, python
from .types.primitives import (
    ALOAD_name, ASTORE_name, free_name,
)
# from .debug import DEBUG, DEBUG_value


class Class(Block):
    CONSTRUCTOR = InitMethod

    def __init__(
            self, parent, name,
            namespace=None, extends=None, implements=None,
            public=True, final=False, methods=None, fields=None, init=None,
            verbosity=0, include_default_constructor=True):
        super().__init__(parent=parent, verbosity=verbosity)
        self.name = name
        if namespace is None:
            self.namespace = '.'.join([self.module.namespace, self.module.name])
        else:
            self.namespace = namespace

        self._extends = extends

        self.implements = implements if implements else []
        self.public = public
        self.final = final
        self.methods = methods if methods else []
        self.fields = fields if fields else {}
        self.init = init

        self.include_default_constructor = include_default_constructor

        # Track constructors when they are added
        self.init_method = None

        # Mark this class as being a VOC generated class.
        self.fields["__VOC__"] = "Lorg/python/Object;"

        self._constructor = None

        # Store the Module object as a local variable
        self.store_module()

    @property
    def descriptor(self):
        return self._parent.build_child_class_descriptor(self.name)

    @property
    def class_name(self):
        return '.'.join([self._parent.class_name, self.name])

    @property
    def module(self):
        return self._parent.module

    def store_module(self):
        # Stores the current module as a local variable
        if ('#module') not in self.local_vars:
            self.add_opcodes(
                JavaOpcodes.GETSTATIC('python/sys', 'modules', 'Lorg/python/types/Dict;'),

                python.Str(self.module.full_name),

                python.Object.get_item(),
                JavaOpcodes.CHECKCAST('org/python/types/Module'),

                ASTORE_name('#module'),
            )

    def build_child_class_descriptor(self, child_name):
        return '$'.join([self.descriptor, child_name])

    def add_class(self, class_name, extends, implements):
        klass = Class(
            self,
            name=class_name,
            extends=extends,
            implements=implements,
        )

        self.module.classes.append(klass)

        self.add_opcodes(
            # Stack contains the bases list
            ASTORE_name('#bases'),

            # DEBUG("FORCE LOAD OF CLASS %s AT DEFINITION" % klass.descriptor),
            # - class
            JavaOpcodes.LDC_W(JavaClassref(klass.descriptor)),

            # - name
            JavaOpcodes.LDC_W(klass.name),

            # - bases
            ALOAD_name('#bases'),

            # - dict
            JavaOpcodes.ACONST_NULL(),

            JavaOpcodes.INVOKESTATIC(
                'org/python/types/Type',
                'declarePythonType',
                args=[
                    'Ljava/lang/Class;',
                    'Ljava/lang/String;',
                    'Ljava/util/List;',
                    'Ljava/util/Map;'
                ],
                returns='Lorg/python/types/Type;'
            ),

            free_name('#bases')
        )

        self.store_name(klass.name)

        self.add_opcodes(
            JavaOpcodes.INVOKESTATIC(
                klass.descriptor,
                'class$init',
                args=[],
                returns='V'
            ),
        )

        return klass

    def visitor_setup(self):
        self.add_opcodes(
            # DEBUG("DEFINITION BLOCK OF " + self.descriptor),
            # Force the loading and instantiation of the python module
            # that defines the class.
            # This may already be imported, but just in case
            JavaOpcodes.LDC_W(self.module.full_name),
            JavaOpcodes.ACONST_NULL(),
            JavaOpcodes.ACONST_NULL(),
            JavaOpcodes.ACONST_NULL(),
            JavaOpcodes.ICONST_0(),
            JavaOpcodes.INVOKESTATIC(
                'org/python/ImportLib',
                '__import__',
                args=[
                    'Ljava/lang/String;',
                    'Ljava/util/Map;',
                    'Ljava/util/Map;',
                    '[Ljava/lang/String;',
                    'I',
                ],
                returns='Lorg/python/types/Module;'
            ),
            JavaOpcodes.POP(),

            # Set the module name
            python.Str(self.module.full_name),
        )

        self.store_name('__module__')

        self.add_opcodes(
            python.Str(self.name),
        )
        self.store_name('__qualname__')

        # self.add_opcodes(
        #     DEBUG("DEFINITION BLOCK OF " + self.descriptor + " DONE"),
        # )

    def store_name(self, name, declare=False):
        self.add_opcodes(
            ASTORE_name('#value'),
            python.Type.for_class(self.descriptor),

            ALOAD_name('#value'),
            python.Object.set_attr(name),

            free_name('#value')
        )

    def store_dynamic(self):
        self.add_opcodes(
            ASTORE_name('#value'),
            python.Type.for_class(self.descriptor),

            JavaOpcodes.GETFIELD('org/python/types/Type', '__dict__', 'Ljava/util/Map;'),
            ALOAD_name('#value'),

            java.Map.putAll(),
            free_name('#value')
        )

    def load_name(self, name):
        self.add_opcodes(
            python.Type.for_class(self.descriptor),
            python.Object.get_attribute(name),
        )

    def load_globals(self):
        self.add_opcodes(
            ALOAD_name('#module'),
            JavaOpcodes.GETFIELD('org/python/types/Module', '__dict__', 'Ljava/util/Map;'),
        )

    def load_locals(self):
        self.load_globals()

    def load_vars(self):
        self.load_globals()

    def delete_name(self, name):
        self.add_opcodes(
            python.Type.for_name(self.descriptor),
            python.Object.del_attr(name),
        )

    @property
    def constructor(self):
        # Make sure there is a Java constructor
        if self._constructor is None:
            init = self.CONSTRUCTOR(self)
            init.visitor_setup()
            init.visitor_teardown()
            return init
        return self._constructor

    @constructor.setter
    def constructor(self, value):
        self._constructor = value

    def add_function(self, name, code, parameter_signatures, return_signature):
        if code.co_flags & CO_GENERATOR:
            method = GeneratorMethod(
                self,
                name=name,
                code=code,
                generator=code.co_name,
                parameters=parameter_signatures,
                returns=return_signature,
                static=True
            )

        else:
            method = Method(
                self,
                name=name,
                code=code,
                parameters=parameter_signatures,
                returns=return_signature,
                static=True
            )

        # Add the method to the list that need to be
        # transpiled into Java methods
        self.methods.append(method)

        # Add a definition of the callable object
        self.add_callable(method)

        if method.name == '__init__':
            self.init_method = method

        return method

    def visitor_teardown(self):
        self.add_opcodes(
            JavaOpcodes.RETURN()
        )

    @property
    def extends(self):
        if self._extends:
            return self._extends
        else:
            return 'org.python.types.Object'

    @property
    def extends_descriptor(self):
        return self.extends.replace('.', '/')

    def transpile(self):
        classfile = JavaClass(
            self.descriptor,
            extends=self.extends_descriptor,
            implements=self.implements,
            public=self.public,
            final=self.final
        )

        classfile.attributes.append(
            SourceFile(os.path.basename(self.module.sourcefile))
        )

        classfile.attributes.append(
            RuntimeVisibleAnnotations([
                Annotation(
                    'Lorg/python/Method;',
                    {
                        '__doc__': ConstantElementValue("Python Class (insert docs here)")
                    }
                )
            ])
        )

        try:
            # If we have block content, add a static block to the class
            static_init = JavaMethod('class$init', '()V', public=False, static=True)
            static_init.attributes.append(super().transpile())
            classfile.methods.append(static_init)
        except IgnoreBlock:
            pass

        # Add any manually defined fields
        classfile.fields.extend([
            JavaField(name, descriptor)
            for name, descriptor in self.fields.items()
        ])

        # Add any methods
        for method in self.methods:
            classfile.methods.extend(method.transpile())

        # Add the constructor
        classfile.methods.extend(self.constructor.transpile())

        # Ensure the class has a class protected, no-args init() so that
        # instances can be instantiated.
        if self.include_default_constructor:
            classfile.methods.append(
                JavaMethod(
                    '<init>',
                    '()V',
                    public=False,
                    static=False,
                    attributes=[
                        JavaCode(
                            max_stack=1,
                            max_locals=1,
                            code=[
                                JavaOpcodes.ALOAD_0(),
                                JavaOpcodes.INVOKESPECIAL(self.extends_descriptor, '<init>', args=[], returns='V'),
                                JavaOpcodes.RETURN(),
                            ]
                        )
                    ]
                )
            )

        return self.namespace, self.descriptor.split('/')[-1], classfile


class ClosureClass(Class):
    CONSTRUCTOR = ClosureInitMethod

    def __init__(self, parent, name, verbosity=0):
        super().__init__(
            parent=parent,
            name=name,
            extends='org/python/types/Closure',
            implements=['org/python/Callable'],
            verbosity=verbosity,
            include_default_constructor=False,
        )



from ..utils import TranspileTestCase


class AssertTests(TranspileTestCase):
    def test_assert(self):
        self.assertCodeExecution("""
            x = 10
            print("Pre assert 1")
            assert x > 0, "It's big"
            print("Pre assert 2")
            try:
                assert x < 0, "It isn't big"
            except AssertionError as err:
                print(err)
            print("Done.")
            """)

    def test_assert_without_message(self):
        self.assertCodeExecution("""
            x = 10
            print("Pre assert 1")
            assert x > 0
            print("Pre assert 2")
            try:
                assert x < 0
            except AssertionError as err:
                print(err)
            print("Done.")
            """)
